import sys
import pathlib
working_directory = pathlib.Path().absolute()
script_directory = pathlib.Path(__file__).parent.absolute()
sys.path.append(str(working_directory))
import unittest
from test_dataset_base import TestDatasetBase
from test_preprocessing import TestPreprocessingBase


from pyadlml.dataset import set_data_home
from pyadlml.dataset import ACTIVITY, DEVICE, END_TIME, START_TIME, VAL, TIME,\
    set_data_home, fetch_amsterdam, clear_data_home
TEST_DATA_HOME = '/tmp/pyadlml_testing'

class TestAmsterdamDataset(TestDatasetBase):
    __test__ = True
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fetch_method = fetch_amsterdam
        self.data_home = TEST_DATA_HOME
        self.df_activity_attrs = ['df_activities']
        self.lst_activity_attrs = ['lst_activities']
        self.num_activities = [7]
        self.len_activities = [263]
        self.num_rec_acts = [7]

        self.len_devices = 2620
        self.num_devices = 14
        self.num_rec_devs = 14
        self.num_rec_binary_devs = 14

    def _setUp(self):
        set_data_home(TEST_DATA_HOME)
        self.data = fetch_amsterdam(keep_original=True, cache=False)

    def test_state_vector_encoder(self):
        from pyadlml.preprocessing import StateVectorEncoder
        df_dev = self.data.df_devices

        # test state vector encoder
        sve = StateVectorEncoder(encode='raw')
        x = sve.fit_transform(df_dev)

        sve = StateVectorEncoder(encode='changepoint')
        x = sve.fit_transform(df_dev)

        sve = StateVectorEncoder(encode='last_fired')
        x = sve.fit_transform(df_dev)


class TestAmsterdamPreprocessing(TestPreprocessingBase):
    __test__ = True
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fetch_method = fetch_amsterdam
        self.data_home = TEST_DATA_HOME
        self.df_activity_attrs = ['df_activities']
        self.lst_activity_attrs = ['lst_activities']


if __name__ == '__main__':
    unittest.main()


import os
import multiprocessing

ENV_PARALLEL='PYADLML_PARALLEL'

def set_parallel(val):
    """ tells code to run execute tasks with dask in parallel whenever possible
    """
    assert isinstance(val, bool)
    os.environ[ENV_PARALLEL] = str(val)
    assert get_parallel() == val

def get_parallel():
    val = os.environ.get(ENV_PARALLEL)
    assert val in ["True", "False"]
    return val == "True"
 	
def get_npartitions():
    """ returns the num of parallel threads that can work on a computer
    """
    return 4*multiprocessing.cpu_count()


import os

ENV_PRIMARY_COLOR='PYADLML_PRIMARY_COLOR'
ENV_SECONDARY_COLOR='PYADLML_SECONDARY_COLOR'
ENV_DIV_COLORMAP='PYADLML_DIVERGING_COLOR'
ENV_SEQ_COLORMAP='PYADLML_SEQUENTIAL_COLOR'
# theming
def set_primary_color(val):
    os.environ[ENV_PRIMARY_COLOR] = val

def set_secondary_color(val):
    os.environ[ENV_SECONDARY_COLOR] = val

def set_diverging_color(val):
    os.environ[ENV_DIV_COLORMAP] = val

def set_sequential_color(val):
    os.environ[ENV_SEQ_COLORMAP] = val

def get_primary_color():
    return os.environ[ENV_PRIMARY_COLOR]

def get_secondary_color():
    return os.environ[ENV_SECONDARY_COLOR]

def get_diverging_color():
    return os.environ[ENV_DIV_COLORMAP]

def get_sequential_color():
    return os.environ[ENV_SEQ_COLORMAP]


import numpy as np
import pandas as pd


"""

metrics oriented at torch.metrics

"""

def online_accuracy(y_true: np.ndarray, y_pred: np.ndarray, times: np.ndarray, n_classes: int, average: str = None):
    """

    Parameters
    ----------

    average: str one of ['micro', 'macro', 'weighted']

    """
    assert average in [None, 'micro', 'macro', 'weighted']

    # An array where for each class the specified quantities are computed

    if average == 'micro':
        df_confmat = online_confusion_matrix(y_true, y_pred, times, n_classes)
        confmat = df_confmat.values
        tp = np.diag(confmat)

        preds = y_pred.flatten()
        target = y_true.flatten()
        tp = (preds == target).sum()

        """
        Normally:
            A FP for class i counts when i is not (j), but i was predicted (j!=i)
            A FN for class i counts when i is but j was predicted (j!=i)
            If there is a FP for class i it counts as FN for class j and vice versa,
            therefore FP = FN in multiclass settings
        Since each sample has a weight, a specific duration a FN 
        """
        fp = (preds != target).sum()
        fn = (preds != target).sum()
        #tn = n_classes * preds.numel() - (fp + fn + tp)
        return tp.sum() / (tp.sum() + fn.sum())

    elif average == 'macro':
        df_confmat = online_confusion_matrix(y_true, y_pred, times, n_classes)

        # Extract true as columns and pred as rows in nanosecond resolution
        # c_ij = predicted class i when true class is j
        confmat = df_confmat.values
        tp = np.diag(confmat)

        # Sum over is sth. else but was mistaken as c
        fp = confmat.sum(1) - tp

        # Sum over is c but was predicted as sth. else
        # recovers y_true with df.groupby('y_true')['diff'].sum()
        fn = confmat.sum(0) - tp

        # All predictions that do not include c
        tn = confmat.sum() - (fp + fn + tp)

        # Array of per class precisions? 
        score =  tp / (tp + fn)

        # Normalize by each class by its 
        normalized_score = (score*(1/n_classes)).sum(-1)
        return normalized_score


def online_confusion_matrix(y_true: np.ndarray, y_pred: np.ndarray, times: np.ndarray, 
                            n_classes: int, average=''):
    """
    Rows are predictions and columns are true values
    c_ij = predicted class is i at i-th row and true class j

    Returns
    -------
    cm : pd.DataFrame

    """
    # TODO refactor, check why y_true has shape (N, 1)
    y_true, y_pred, times = y_true.squeeze(), y_pred.squeeze(), times.squeeze()

    df = pd.DataFrame(data=[y_true, y_pred, times], index=['y_true', 'y_pred', 'times']).T
    df['times'] = pd.to_datetime(df['times'])
    df['diff'] = df['times'].shift(-1) - df['times']

    # Impute last unknown time difference with mean
    last_activity = df.at[df.shape[0]-1, 'y_true']
    last_act_mean = df.groupby(y_true).mean().at[last_activity, 'diff']
    df.at[df.shape[0]-1, 'diff'] = last_act_mean

    cm = pd.crosstab(index=df['y_pred'], columns=df['y_true'], values=df['diff'], aggfunc=np.sum)
    cm = cm.replace({pd.NaT: pd.Timedelta('0s')})

    missing_row = set(cm.columns) - set(cm.index)
    missing_col = set(cm.index) - set(cm.columns)
    if missing_row:
        # y_pred did not contain classes present in y_true
        for m_row in missing_row:
            new_row = pd.Series(name=m_row, 
                                data=[pd.Timedelta('0s')]*len(cm.columns), 
                                index=cm.columns
            )
            cm = pd.concat([cm, new_row.to_frame().T], axis=0)

    if missing_col:
        # y_true did not contain classes present in y_pred
        for m_col in missing_col:
            new_col = pd.Series(name=m_col, 
                                data=[pd.Timedelta('0s')]*len(cm.rows), 
                                index=cm.rows
            )
            cm = pd.concat([cm, new_col.to_frame().T], axis=1)

        # TODO if this happens check if it works
        raise NotImplementedError

    # Rearange columns and for diagonal to again matchup 
    # when col or row was inserted above
    cm = cm.sort_index(ascending=True)
    cm = cm[cm.index.values]

    return cm 


from pathlib import Path
import numpy as np
from sklearn.preprocessing import LabelEncoder
from pyadlml.constants import DEVICE, TIME, VALUE
from torch.utils.data import DataLoader
from pyadlml.metrics import online_accuracy
from pyadlml.model.plot.mp import plot_activation_dist, plot_gradient_dist, plot_weight_gradient_dist
from pyadlml.model.transformer.vanilla import VanillaTransformer

from pyadlml.model_selection import train_test_split
from pyadlml.dataset import *
from pyadlml.preprocessing import IndexEncoder, DropColumn, StateVectorEncoder, LabelMatcher, DropTimeIndex, \
    DropDuplicates, EventWindow
from pyadlml.pipeline import EvalOnlyWrapper, Pipeline, TrainOnlyWrapper
from sklearn.ensemble import RandomForestClassifier
from ray import tune
from pyadlml.model import MLP, WaveNet
from pyadlml.dataset.torch import TorchDataset
from pyadlml.preprocessing.preprocessing import KeepOnlyDevices
from pyadlml.training.trainable import Trainable 
import mlflow
import tempfile
import argparse
from pyadlml.plot import plotly_activities_and_devices
import torch
from torchmetrics.classification import MulticlassAccuracy, MulticlassConfusionMatrix
from torch.nn.functional import softmax
from pyadlml.preprocessing import Timestamp2Seqtime

DEVICE = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

"""
File for developing and debugging models

Dataset avaliable:
    'casas_aruba',
    'amsterdam',
    'kasteren_A',
    'uci_ordonezA',
    'uci_ordonezB',

n    'mitlab_1',
n    'mitlab_2',
n    'kasteren_B',
n    'kasteren_C',
"""


def main():
    parser=argparse.ArgumentParser(description="train_debug tool")
    parser.add_argument("dataset", type=str)
    args=parser.parse_args()

    seed = 1
    dataset =  args.dataset


    mlflow.set_tracking_uri("http://localhost:5555")

    class TrainableDebug(Trainable):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        

        def __call__(self, hparam_dict):
            """
            """
            from sklearn.base import clone

            # Clone pipe and initialize with values
            pipe_params = hparam_dict.pop('pipe_params')
            pipe = clone(self.pipe)
            pipe.set_params(**pipe_params)

            with tempfile.TemporaryDirectory() as workdir:
                workdir = Path(workdir)
                model_name = hparam_dict['model']
                run_id = self._get_run_id_of(model_name)

                with mlflow.start_run(run_id, self.exp.experiment_id, model_name) as run:
                    # Create new trial
                    with mlflow.start_run(None, run.info.experiment_id, self._create_trial_name(run.info.run_id), True) as sub_run:
                        
                        mlflow.log_params(hparam_dict)
                        mlflow.log_params(pipe_params)
                        mlflow.set_tags({'dataset': self.ds_name,'debugging': True})

                        # Fetch data and do train test split
                        df_devs, df_acts = self._fetch_dataset()
                        X_train, X_val, X_test, y_train, y_val, y_test, init_states = train_test_split(
                            df_devs, df_acts, split=(0.6, 0.2, 0.2), temporal=True,
                            return_init_states=True
                        )


                        # Log train/val/test data to mlflow
                        mlflow.log_text(X_train.to_csv(index=False), artifact_file='data/X_train.csv')
                        mlflow.log_text(y_train.to_csv(index=False), artifact_file='data/y_train.csv')
                        mlflow.log_figure(plotly_activities_and_devices(X_train, y_train), artifact_file='data/Xy_train.html')

                        mlflow.log_text(X_val.to_csv(index=False), artifact_file='data/X_val.csv')
                        mlflow.log_text(y_val.to_csv(index=False), artifact_file='data/y_val.csv')
                        mlflow.log_figure(plotly_activities_and_devices(X_val, y_val), artifact_file='data/Xy_val.html')

                        mlflow.log_text(X_test.to_csv(index=False), artifact_file='data/X_test.csv')
                        mlflow.log_text(y_test.to_csv(index=False), artifact_file='data/y_test.csv')
                        mlflow.log_figure(plotly_activities_and_devices(X_test, y_test), artifact_file='data/Xy_test.html')


                        pipe.train()
                        train_loader = DataLoader(
                            dataset=TorchDataset(X_train, y_train, transforms=pipe),
                            batch_size=hparam_dict['batch_size'],
                            shuffle=True,
                            drop_last=True      # BatchNorm1d not working otherwise
                        )

                        pipe.eval()
                        # Validate pipeline
                        val_loader = DataLoader(
                            dataset=TorchDataset(X_val, y_val, transforms=pipe, 
                                # With current representaiton not used
                                #transform_params=dict(enc__initial_states=init_states['init_states_val'])
                            ),
                            batch_size=hparam_dict['batch_size'],
                            shuffle=False,
                            drop_last=True  # BatchNorm1d not working otherwise
                        )
                        n_classes=len(train_loader.dataset.classes_)

                        #model = MLP(
                        #    n_features=12,  # Pipe returns 12 devices
                        #    n_hidden=hparam_dict['dim_hidden'],
                        #    window_length=pipe_params['time_windows__window_size'],
                        #    n_embed=hparam_dict['dim_emb'],
                        #    n_classes= len(train_loader.dataset.classes_),
                        #)
                        model = WaveNet(
                            n_features=12, 
                            n_emb=hparam_dict['dim_emb'],
                            n_hidden=hparam_dict['dim_hidden'],
                            n_classes=len(train_loader.dataset.classes_),
                            dilation_factor=hparam_dict['dilation_factor']
                        ).to(DEVICE)
                        #from pyadlml.model.transformer.vanilla import VanillaTransformer
                        #model = VanillaTransformer(
                        #    n_embd=hparam_dict['dim_embd'],
                        #    n_features=12,
                        #    dropout=hparam_dict['dropout'],
                        #    block_size=pipe_params['windows__window_size'],
                        #    bias=hparam_dict['bias'],           # bias in lin and layernorms
                        #    n_classes=n_classes,
                        #    n_layer=hparam_dict['n_layer'],      # Sequential attention layers
                        #    n_head=hparam_dict['n_heads'],       # Parallel attention blocks attn layers
                        #)
                        #from pyadlml.model.tpp import SAHP
                        #model = SAHP(
                        #    nLayers=hparam_dict['n_layer'],     # Successive attention blocks
                        #    d_model=hparam_dict['dim_embd'],     # Embedding dimensions
                        #    atten_heads=hparam_dict['n_heads'], # Parallel attention heads per block
                        #    dropout=hparam_dict['dropout'],
                        #    process_dim=1,                      # Nr devices
                        #    device=DEVICE,
                        #    max_seq_len=pipe_params['windows__window_size'],

                        #).to(DEVICE)

                        parameters = model.parameters()
                        optimizer = torch.optim.Adam(parameters, 
                            lr=hparam_dict['initial_lr'],
                            betas=(0.9, 0.98),
                            eps=1e-9,
                            weight_decay=3e-4)
                        #optimizer = torch.optim.SGD(
                        #    parameters, 
                        #    lr=hparam_dict['initial_lr']
                        #)

                        B, T, window_type = hparam_dict['batch_size'], pipe_params['windows__window_size'], pipe_params['windows__rep']

                        # Simple training routine
                        max_epochs = 2
                        from torch.nn import functional as F
                        class_weights = train_loader.dataset.class_weights.to(DEVICE)

                        # Initialize metrics
                        n_classes = len(train_loader.dataset.classes_)
                        val_acc_macro = MulticlassAccuracy(num_classes=n_classes, average='macro')
                        val_acc_micro = MulticlassAccuracy(num_classes=n_classes, average='micro')

                        from tqdm import tqdm
                        val_y_trues = []
                        val_y_preds = []

                        for epoch in range(0, max_epochs):
                            with tqdm(train_loader, unit='batch') as t_epoch:
                                model.train()

                                for (Xb, yb) in t_epoch:
                                    Xb, yb = Xb.to(DEVICE), yb.to(DEVICE)
                                    t_epoch.set_description(f"Epoch {epoch}")

                                    optimizer.zero_grad()

                                    # forward pass
                                    logits = model(Xb)
                                    loss = F.cross_entropy(logits.view(-1, logits.shape[-1]), yb.view(-1)) 
                                    #loss = model.compute_loss(Xb, n_mc_samples=100)
                                    # Case classification
                                    #loss = model.compute_loss(
                                    #    logits.view(-1, logits.shape[-1]),
                                    #    yb.view(-1),
                                    #    weight=class_weights
                                    #)
                                    
                                    loss.backward()
                                    optimizer.step()
                                    

                                    t_epoch.set_postfix(train_loss=loss.item())
                                    mlflow.log_metric('train_loss', loss.item())

                                Y_true = []
                                Y_conf = []
                                Y_pred = []
                                with torch.no_grad():
                                    model.eval()
                                    for (Xb, yb) in val_loader:
                                        Xb, yb = Xb.to(DEVICE), yb.to(DEVICE)
                                        logits = model(Xb)  # B, T, C
                                        
                                        # Expects N, C, therefore flatten logits 
                                        loss = F.cross_entropy(logits.view(-1, logits.shape[-1]), yb.view(-1)) 

                                        # (B, T, C) -> (B, T)
                                        y_pred = torch.argmax(logits, dim=-1).detach().cpu()
                                        # (B, T, C) -> (B, T, C)
                                        y_conf = softmax(logits, dim=-1).detach().cpu()
                                        # (B, T) -> (B, T)
                                        y_true_times = yb.detach().cpu()

                                        Y_true.append(y_true_times)
                                        Y_pred.append(y_pred)
                                        Y_conf.append(y_conf)

                                        acc_ma = val_acc_macro(y_true_times, y_pred)
                                        acc_mi = val_acc_micro(y_true_times, y_pred)

                                        mlflow.log_metric('val_loss', loss.item())
                                        t_epoch.set_postfix(val_loss=loss.item())

                                    val_acc_macro.compute()
                                    val_acc_micro.compute()
                                    mlflow.log_metric('val_acc_micro', acc_mi)
                                    mlflow.log_metric('val_acc_macro', acc_ma)

                                mlflow.log_metric('epoch', epoch)

                                Y_true = torch.cat(Y_true, dim=0)
                                Y_pred = torch.cat(Y_pred, dim=0)
                                if window_type == 'many-to-many':
                                    # TODO refactor do without if
                                    #Y_true has shape (B*n, T) -> flatten
                                    Y_true = Y_true.flatten()
                                    Y_pred = Y_pred.flatten()
                                Y_conf = torch.cat(Y_conf, dim=0)
                                val_y_trues.append(Y_true)
                                val_y_preds.append(Y_pred)


                        # Post training
                        #mlflow.log_figure(plot_activation_dist(model.layers), 
                        #                  'eval/activation_dist.png')
                        #mlflow.log_figure(plot_gradient_dist(model.layers), 
                        #                  'eval/gradient_dist.png') 
                        #mlflow.log_figure(plot_weight_gradient_dist(parameters), 
                        #                  'eval/weight_gradient_dist.png')
                        from pyadlml.ml_viz import plotly_confusion_matrix
                        fig = plotly_confusion_matrix(val_y_preds, val_y_trues, val_loader.dataset.classes_, per_step=True)
                        mlflow.log_figure(fig, 'eval/cm.html')

                        # Retrieve time points where y is predicted from pipe
                        times_prae_win = pipe.get_times()
                        y_true_times = pipe['windows'].construct_target_times(times_prae_win)

                        # Recreate according 
                        Xt_prae_win, _ = pipe[:-2].transform(X_val, y_val)

                        devs = pipe['windows'].construct_X_at_target(Xt_prae_win)

                        # One hot encode
                        onehot = pipe['enc'].onehot(devs)

                        if window_type == 'many-to-many':
                            # Restore sequence length dimension since (B,T) -> (n*B,T) -> (n*B*T) where n is val-loader steps
                            # (n*B*T) -> (n*B,T) -> (n*B, 1) -> (n*B)Then get last elemnt of sequence1
                            Y_true = Y_true.view(-1, T)[:, [-1]].squeeze()
                            Y_pred = Y_pred.view(-1, T)[:, [-1]].squeeze()
                            # (B, T, C) -> (B, 1, C)
                            Y_conf = Y_conf[:, [-1], :].squeeze(1)

                        # Correct for last batch dropped 
                        if len(y_true_times) > Y_true.shape[0]:
                            offset = -(len(y_true_times) - Y_true.shape[0])
                            y_true_times = y_true_times[:offset]
                            onehot = onehot[:offset]

                        assert y_true_times.shape[0] == onehot.shape[0]


                        y_conf = Y_conf.detach().numpy()
                        i2c = np.vectorize(val_loader.dataset.int2class)
                        y_pred = i2c(Y_pred.detach().numpy())
                        y_true = i2c(Y_true.detach().numpy())

                        classes_ = val_loader.dataset.classes_

                        from pyadlml.dataset.plot.plotly.discrete import acts_and_devs
                        f_and = acts_and_devs(onehot, y_true, y_pred, y_conf, classes_)
                        mlflow.log_figure(f_and, artifact_file='eval/val_predictions.html')
                        f_and = acts_and_devs(onehot, y_true, y_pred, y_conf, classes_, y_true_times)
                        mlflow.log_figure(f_and, artifact_file='eval/val_predictions_time.html')


                        # Compute additional online accuracies
                        print()
                        online_acc_macro = online_accuracy(y_true, y_pred, y_true_times, n_classes=len(classes_), average='macro')
                        online_acc_micro = online_accuracy(y_true, y_pred, y_true_times, n_classes=len(classes_), average='micro')
                        mlflow.log_metric('val_online_acc_micro', online_acc_micro)
                        mlflow.log_metric('val_online_acc_macro', online_acc_macro)








    trainable = TrainableDebug(exp_name='masterarbeit', ds_name=dataset)


    #    model='Self Attentive Hawkes Process',
    hp_pipe = Pipeline([
            ('enc', IndexEncoder()),
            ('drop_obs', DropColumn(VALUE)),
            ('lbl', LabelMatcher()),
            ('drop_time', DropTimeIndex(only_y=True)),
            ('ts2seq', Timestamp2Seqtime()),
            ('windows', EventWindow()),
            ('passthrough', 'passthrough')
            ]
        )
    #hparam_dict = dict(
    #    #model='MLP',
    #    model='Self Attentive Hawkes Process',
    #    dim_embd=16,
    #    batch_size=32,
    #    #initial_lr=5e-5,
    #    initial_lr=6e-5,
    #    bias=False,
    #    n_heads=8,
    #    n_layer=4,
    #    dropout=0.1,
    #    pipe_params=dict(
    #        lbl__other=True,
    #        windows__rep = 'many-to-many',
    #        windows__window_size = 100,
    #    )
    #)



    # Define Hyperparameter and parameters for mlflow
    # 4 layers -> dilation 2 -> 2^4=16
    # 4 layers -> dilation 3 -> 3^4=81
    # (dim, win, h) -> (2, 16, 68)
    # (dim, win, h) -> (3, 27, 68)
    hparam_dict = dict(
        model='Wavenet',
        batch_size=32,
        dim_emb=4,
        dilation_factor=2,
        dim_hidden=200,
        initial_lr=1e-3,
        bias=False,
        pipe_params=dict(
            lbl__other=True,
            windows__rep = 'many-to-one',
            windows__window_size = 16,
        )
    )

    trainable.set_pipe(
        Pipeline([
            ('enc', IndexEncoder()),
            ('drop_obs', DropColumn(VALUE)),
            ('lbl', LabelMatcher()),
            ('drop_time', DropTimeIndex()),
            ('windows', EventWindow()),
            ('passthrough', 'passthrough')
            ]
        )
    )

    torch.backends.cudnn.allow_tf32 = True # allow tf32 on cudnn
    torch.backends.cuda.allow_tf32 = True # allow tf32 on cudnn
    torch.manual_seed(1337)

    # initialize a GradScaler. If enabled=False scaler is a no-op
    #scaler = torch.cuda.amp.GradScaler(enabled=(dtype == 'float17'))


    trainable(hparam_dict)

if __name__ == '__main__':
    main()


from pyadlml.constants import  TIME, START_TIME, END_TIME, ACTIVITY, DEVICE
from pyadlml.dataset._core.activities import ActivityDict
from pyadlml.dataset.stats.activities import coverage
from pyadlml.dataset.util import infer_dtypes, fetch_by_name, DATASET_STRINGS
import argparse

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--dataset', type=str, default='amsterdam',

                        choices=DATASET_STRINGS,
                        help='Select an avaliable dataset.'
                        )
    parser.add_argument('-i', '--identifier', type=str, default=None,
                        help='Specify dataset specific stuff'
    )
    args = parser.parse_args()

    data = fetch_by_name(args.dataset, args.identifier, cache=False)

    df_devs = data['devices']
    dct_acts = ActivityDict(data['activities'])

    # Determine start and end_time
    start_time = min(df_devs[TIME].iloc[0], dct_acts.min_starttime())
    end_time = max(df_devs[TIME].iloc[-1], dct_acts.max_endtime())
    # Compute device statistics
    nr_devices = len(df_devs[DEVICE].unique())
    nr_device_recordings = len(df_devs)
    dev_types = [k for k, v in infer_dtypes(df_devs).items() if len(v) > 0]


    residents = {r:{} for r in dct_acts.keys()}


    print(f':From: {start_time}')
    print(f':To: {end_time}')
    print(f':Devices: {nr_devices}/{nr_device_recordings}')
    print(f':DeviceType: {str(dev_types)}')

    for r in residents.keys():
        df_acts = dct_acts[r]

        # Compute activity statistics
        nr_activities = len(df_acts[ACTIVITY].unique())
        nr_activity_recordings = len(df_acts)

        act_cov_dp = coverage(df_acts, df_devs, datapoints=True)
        act_cov_time = coverage(df_acts, df_devs, datapoints=False)

        print(f":Activites {r}:\t{nr_activities}/{nr_activity_recordings}")
        print(f":Coverage {r}:\t{act_cov_time:.2f}/{act_cov_dp:.2f}")



import math
import torch
import torch.nn as nn
import torch.utils
import torch.utils.data
from torchvision import datasets, transforms
from torch.autograd import Variable
import matplotlib.pyplot as plt


"""implementation of the Variational Recurrent
Neural Network (VRNN) from https://arxiv.org/abs/1506.02216
using unimodal isotropic gaussian distributions for 
inference, prior, and generating models."""


class VRNN(nn.Module):
	def __init__(self, x_dim, h_dim, z_dim, n_layers, bias=False):
		super(VRNN, self).__init__()

		self.x_dim = x_dim
		self.h_dim = h_dim
		self.z_dim = z_dim
		self.n_layers = n_layers

		# feature-extracting transformations
		self.phi_x = nn.Sequential(
			nn.Linear(x_dim, h_dim),
			nn.ReLU(),
			nn.Linear(h_dim, h_dim),
			nn.ReLU())
		self.phi_z = nn.Sequential(
			nn.Linear(z_dim, h_dim),
			nn.ReLU())

		# encoder
		self.enc = nn.Sequential(
			nn.Linear(h_dim + h_dim, h_dim),
			nn.ReLU(),
			nn.Linear(h_dim, h_dim),
			nn.ReLU())
		self.enc_mean = nn.Linear(h_dim, z_dim)
		self.enc_std = nn.Sequential(
			nn.Linear(h_dim, z_dim),
			nn.Softplus())

		# prior
		self.prior = nn.Sequential(
			nn.Linear(h_dim, h_dim),
			nn.ReLU())
		self.prior_mean = nn.Linear(h_dim, z_dim)
		self.prior_std = nn.Sequential(
			nn.Linear(h_dim, z_dim),
			nn.Softplus())

		# decoder
		self.dec = nn.Sequential(
			nn.Linear(h_dim + h_dim, h_dim),
			nn.ReLU(),
			nn.Linear(h_dim, h_dim),
			nn.ReLU())
		self.dec_std = nn.Sequential(
			nn.Linear(h_dim, x_dim),
			nn.Softplus())
		#  self.dec_mean = nn.Linear(h_dim, x_dim)
		self.dec_mean = nn.Sequential(
			nn.Linear(h_dim, x_dim),
			nn.Sigmoid())

		# recurrence
		self.rnn = nn.GRU(h_dim + h_dim, h_dim, n_layers, bias)


	def forward(self, x):

		all_enc_mean, all_enc_std = [], []
		all_dec_mean, all_dec_std = [], []
		kld_loss = 0
		nll_loss = 0

		h = Variable(torch.zeros(self.n_layers, x.size(1), self.h_dim))
		for t in range(x.size(0)):

			phi_x_t = self.phi_x(x[t])

			# encoder
			enc_t = self.enc(torch.cat([phi_x_t, h[-1]], 1))
			enc_mean_t = self.enc_mean(enc_t)
			enc_std_t = self.enc_std(enc_t)

			# prior
			prior_t = self.prior(h[-1])
			prior_mean_t = self.prior_mean(prior_t)
			prior_std_t = self.prior_std(prior_t)

			# sampling and reparameterization
			z_t = self._reparameterized_sample(enc_mean_t, enc_std_t)
			phi_z_t = self.phi_z(z_t)

			# decoder
			dec_t = self.dec(torch.cat([phi_z_t, h[-1]], 1))
			dec_mean_t = self.dec_mean(dec_t)
			dec_std_t = self.dec_std(dec_t)

			# recurrence
			_, h = self.rnn(torch.cat([phi_x_t, phi_z_t], 1).unsqueeze(0), h)

			# computing losses
			kld_loss += self._kld_gauss(enc_mean_t, enc_std_t, prior_mean_t, prior_std_t)
			# nll_loss += self._nll_gauss(dec_mean_t, dec_std_t, x[t])
			nll_loss += self._nll_bernoulli(dec_mean_t, x[t])

			all_enc_std.append(enc_std_t)
			all_enc_mean.append(enc_mean_t)
			all_dec_mean.append(dec_mean_t)
			all_dec_std.append(dec_std_t)

		return kld_loss, nll_loss, \
			(all_enc_mean, all_enc_std), \
			(all_dec_mean, all_dec_std)


	def sample(self, seq_len):

		sample = torch.zeros(seq_len, self.x_dim)

		h = Variable(torch.zeros(self.n_layers, 1, self.h_dim))
		for t in range(seq_len):

			# prior
			prior_t = self.prior(h[-1])
			prior_mean_t = self.prior_mean(prior_t)
			prior_std_t = self.prior_std(prior_t)

			# sampling and reparameterization
			z_t = self._reparameterized_sample(prior_mean_t, prior_std_t)
			phi_z_t = self.phi_z(z_t)

			# decoder
			dec_t = self.dec(torch.cat([phi_z_t, h[-1]], 1))
			dec_mean_t = self.dec_mean(dec_t)
			# dec_std_t = self.dec_std(dec_t)

			phi_x_t = self.phi_x(dec_mean_t)

			# recurrence
			_, h = self.rnn(torch.cat([phi_x_t, phi_z_t], 1).unsqueeze(0), h)

			sample[t] = dec_mean_t.data_

		return sample


	def reset_parameters(self, stdv=1e-1):
		for weight in self.parameters():
			weight.data_.normal_(0, stdv)


	def _init_weights(self, stdv):
		pass


	def _reparameterized_sample(self, mean, std):
		"""using std to sample"""
		eps = torch.FloatTensor(std.size()).normal_()
		eps = Variable(eps)
		return eps.mul(std).add_(mean)

	def _kld_gauss(self, mean_1, std_1, mean_2, std_2):
		"""Using std to compute KLD"""

		kld_element = (2 * torch.log(std_2) - 2 * torch.log(std_1) +
			(std_1.pow(2) + (mean_1 - mean_2).pow(2)) /
			std_2.pow(2) - 1)
		return 0.5 * torch.sum(kld_element)

	def _nll_bernoulli(self, theta, x):
		return - torch.sum(x*torch.log(theta) + (1-x)*torch.log(1-theta))


	def _nll_gauss(self, mean, std, x):
		pass


from pathlib import Path
import pandas as pd
from sktime.classification.kernel_based import RocketClassifier
import numpy as np
from sklearn.preprocessing import LabelEncoder
from pyadlml.constants import DEVICE, TIME, VALUE
from torch.utils.data import DataLoader
from pyadlml.dataset.plot.plotly.discrete import acts_and_devs
from pyadlml.metrics import online_accuracy
from pyadlml.model.plot.mp import plot_activation_dist, plot_gradient_dist, plot_weight_gradient_dist
from pyadlml.model.transformer.vanilla import VanillaTransformer

from pyadlml.model_selection import train_test_split
from pyadlml.dataset import *
from pyadlml.preprocessing import IndexEncoder, DropColumn, StateVectorEncoder, LabelMatcher, DropTimeIndex, \
    DropDuplicates, EventWindow
from pyadlml.pipeline import EvalOnlyWrapper, Pipeline, TrainOnlyWrapper
from sklearn.ensemble import RandomForestClassifier
from ray import tune
from pyadlml.model import MLP, WaveNet
from pyadlml.dataset.torch import TorchDataset
from pyadlml.preprocessing.preprocessing import KeepOnlyDevices, SkTime
from pyadlml.training.trainable import Trainable 
import mlflow
import tempfile
import argparse
from pyadlml.plot import plotly_activities_and_devices
import torch
from torchmetrics.classification import MulticlassAccuracy, MulticlassConfusionMatrix
from torch.nn.functional import softmax
from pyadlml.preprocessing import Timestamp2Seqtime

DEVICE = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

"""
File for developing and debugging models

Dataset avaliable:
    'casas_aruba',
    'amsterdam',
    'kasteren_A',
    'uci_ordonezA',
    'uci_ordonezB',

n    'mitlab_1',
n    'mitlab_2',
n    'kasteren_B',
n    'kasteren_C',
"""


def main():
    parser=argparse.ArgumentParser(description="train_debug tool")
    parser.add_argument("dataset", type=str)
    args=parser.parse_args()

    seed = 1
    dataset =  args.dataset


    mlflow.set_tracking_uri("http://localhost:5555")

    class TrainableDebug(Trainable):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        

        def __call__(self, hparam_dict):
            """
            """
            from sklearn.base import clone

            # Clone pipe and initialize with values
            pipe_params = hparam_dict.pop('pipe_params')
            pipe = clone(self.pipe)
            pipe.set_params(**pipe_params)

            with tempfile.TemporaryDirectory() as workdir:
                workdir = Path(workdir)
                model_name = hparam_dict['model']
                run_id = self._get_run_id_of(model_name)

                with mlflow.start_run(run_id, self.exp.experiment_id, model_name) as run:
                    # Create new trial
                    with mlflow.start_run(None, run.info.experiment_id, self._create_trial_name(run.info.run_id), True) as sub_run:
                        
                        mlflow.log_params(hparam_dict)
                        mlflow.log_params(pipe_params)
                        mlflow.set_tags({'dataset': self.ds_name,'debugging': True})

                        # Fetch data and do train test split
                        df_devs, df_acts = self._fetch_dataset()
                        X_train, X_val, X_test, y_train, y_val, y_test, init_states = train_test_split(
                            df_devs, df_acts, split=(0.6, 0.2, 0.2), temporal=True,
                            return_init_states=True
                        )


                        # Log train/val/test data to mlflow
                        mlflow.log_text(X_train.to_csv(index=False), artifact_file='data/X_train.csv')
                        mlflow.log_text(y_train.to_csv(index=False), artifact_file='data/y_train.csv')
                        mlflow.log_figure(plotly_activities_and_devices(X_train, y_train), artifact_file='data/Xy_train.html')

                        mlflow.log_text(X_val.to_csv(index=False), artifact_file='data/X_val.csv')
                        mlflow.log_text(y_val.to_csv(index=False), artifact_file='data/y_val.csv')
                        mlflow.log_figure(plotly_activities_and_devices(X_val, y_val), artifact_file='data/Xy_val.html')

                        mlflow.log_text(X_test.to_csv(index=False), artifact_file='data/X_test.csv')
                        mlflow.log_text(y_test.to_csv(index=False), artifact_file='data/y_test.csv')
                        mlflow.log_figure(plotly_activities_and_devices(X_test, y_test), artifact_file='data/Xy_test.html')


                        pipe.train()

                        # Train pipeline
                        pipe.fit(X_train, y_train)

                        # Set into eval mode
                        pipe.eval()



                        # Validate pipeline
                        Xt_val, y_true = pipe[:-1].transform(X_val, y_val, 
                                            enc__initial_states=init_states['init_states_val'])

                        y_times, Xt_val_at_times = pipe.construct_y_times_and_X(X_val, y_val)
                        Xt_val_at_times = pd.DataFrame(Xt_val_at_times, columns=pipe['ev_win'].feature_names_in_)
                        
                        y_pred = pipe[-1].predict(Xt_val) 
                        y_conf = pipe[-1].predict_proba(Xt_val)

                        from sklearn.metrics import accuracy_score
                        acc_score = accuracy_score(y_true, y_pred)

                        n_classes = len(pipe['lbl'].classes_)
                        classes_ = pipe['lbl'].classes_

                        val_acc_macro = MulticlassAccuracy(num_classes=n_classes, average='macro')
                        val_acc_micro = MulticlassAccuracy(num_classes=n_classes, average='micro')
                        lbl_enc = LabelEncoder().fit(y_true)
                        y_true_enc, y_pred_enc = lbl_enc.transform(y_true), lbl_enc.transform(y_pred)
                        acc_mi = val_acc_micro(torch.tensor(y_true_enc), torch.tensor(y_pred_enc))
                        acc_ma = val_acc_macro(torch.tensor(y_true_enc), torch.tensor(y_pred_enc))
                        mlflow.log_metric('val_acc_micro', acc_mi.item())
                        mlflow.log_metric('val_acc_macro', acc_ma.item())

                        from pyadlml.ml_viz import plotly_confusion_matrix
                        fig = plotly_confusion_matrix(y_pred, y_true, classes_)
                        mlflow.log_figure(fig, 'eval/cm.html')

                        #from pyadlml.dataset.plot.plotly.discrete import acts_and_devs
                        f_and = acts_and_devs(Xt_val_at_times, y_true, y_pred, y_conf, classes_)
                        mlflow.log_figure(f_and, artifact_file='eval/val_predictions.html')
                        f_and = acts_and_devs(Xt_val_at_times, y_true, y_pred, y_conf, classes_, y_times)
                        mlflow.log_figure(f_and, artifact_file='eval/val_predictions_time.html')

                        ## Compute additional online accuracies
                        online_acc_macro = online_accuracy(y_true, y_pred, y_times, n_classes=len(classes_), average='macro')
                        online_acc_micro = online_accuracy(y_true, y_pred, y_times, n_classes=len(classes_), average='micro')
                        mlflow.log_metric('val_online_acc_micro', online_acc_micro)
                        mlflow.log_metric('val_online_acc_macro', online_acc_macro)



    trainable = TrainableDebug(exp_name='masterarbeit', ds_name=dataset)

    uci_sklearn_iid_pipe = Pipeline([
        ('enc', StateVectorEncoder()),
        ('lbl', LabelMatcher()),
        ('drop_dups', TrainOnlyWrapper(DropDuplicates())),
        ('drop_time', DropTimeIndex()),
        ('ev_win', EventWindow(rep='many-to-one')),
        ('to_sktime', SkTime(rep='nested_univ')),
        ('clf', RocketClassifier()),
    ])

    hparam_dict = dict(
        model='Rocket',
        pipe_params=dict(
            enc__encode='changepoint',
            lbl__other=True,
            clf__num_kernels=2000,
        )
    )

    trainable.set_pipe(uci_sklearn_iid_pipe)
    trainable(hparam_dict)

if __name__ == '__main__':
    main()


"""
    this is an abstract class of an implemented algorithm like hmm
    model is used for benchmarks
    model is also called from hmm events
"""

from hassbrain_algorithm.controller import Controller
from hassbrain_algorithm.datasets._dataset import DataRep
import joblib
import numpy as np
import copy

MD_FILE_NAME = "model_%s.joblib"

class Model(object):

    def __init__(self, name, controller):
        """

        Parameters
        ----------
        name
        controller

        Attributes
        ----------
        self.__dataset_repr (DataRep)
            is either RAW, LAST_FIRED or CHANGEPOINT
        """
        self._bench = None
        self._ctrl = controller # type: Controller
        self._model_name = MD_FILE_NAME
        self._dataset_repr = DataRep.RAW

        """
        these are callbacks that are can be accessed from third party classes
        for example the Benchmark to execute sth. during each training step.
        The model such as a hmm should call these methods after each step
        """
        self._callbacks = []

        """
        these hashmaps are used to decode and encode in O(1) the numeric based
        values of states and observations to the string based representations 
        """
        self._obs_lbl_hashmap = {}
        self._obs_lbl_rev_hashmap = {}
        self._state_lbl_hashmap = {}
        self._state_lbl_rev_hashmap = {}

        """
        Interpretability
        """
        self._skater_model = None
        self._explanator = None


    def get_data_repr(self):
        return self._dataset_repr

    def get_data_freq(self):
        """
        Returns
        -------
        str
            the frequence of the dataset the model was trained on
            e.g '1min', '30sec', ...
        """
        return self._dataset_conf['freq']

    def can_predict_next_obs(self):
        return NotImplementedError

    def can_predict_prob_devices(self):
        return NotImplementedError

    def encode_state_lbl(self, label):
        return self._state_lbl_hashmap[label]

    def decode_state_lbl(self, ide):
        return self._state_lbl_rev_hashmap[ide]

    def encode_obs_lbl(self, label):
        """
        returns the id of a sensor given a label
        :param label:
        :return:
        """
        return self._obs_lbl_hashmap[label]

    def decode_obs_lbl(self, ide):
        """
        retrieves the label given a sensor id
        :param id:
        :return:
        """
        return self._obs_lbl_rev_hashmap[ide]

    def get_state_lbl_lst(self):
         return list(self._state_lbl_hashmap.keys())

    def get_obs_lbl_lst(self):
        return list(self._obs_lbl_hashmap.keys())

    # todo flag for deletion
    def append_method_to_callbacks(self, callback_method):
        self._callbacks.append(callback_method)

    # todo flag for deletion
    def set_train_loss_callback(self):
        self._callbacks.append(self._train_loss_callback)

    def _train_loss_callback(self, *args):
        """
        hass to format the callback from the real model into an appropriate output for
        the benchmark method bench.train_loss_callback()
        :return:
        """
        raise NotImplementedError

    def register_benchmark(self, bench):
        self._bench = bench


    @classmethod
    def save_model(cls, model, path_to_folder, filename):
        """ saves this instance to a file
        Parameters
        ----------
        path_to_folder
        filename

        Returns
        -------

        """
        full_file_path = path_to_folder + "/" + filename
        import os
        if not os.path.exists(path_to_folder):
            os.makedirs(path_to_folder)
        #import dill
        #dill_file = open(full_file_path, "wb")
        #dill_file.write(dill.dumps(model))
        #dill_file.close()

        joblib.dump(model, full_file_path)

    @classmethod
    def load_model(cls, path_to_folder, filename):
        name = path_to_folder + "/" + filename
        return joblib.load(name)

        #import dill
        #dill_file = open(name, "rb")
        #model = dill.load(dill_file.read())
        #dill_file.close()
        #return model

    def set_name(self, name):
        self._name = name

    def get_name(self):
        return self._name

    def model_init(self, dataset, state_list=None):
        """
        initialize model on dataset
        :param dataset:
        :param state_list:
        :param location_data:
            a list conatining information how the smart home is set up
            loc_data = [ { "name" : "loc1", "activities" : ['cooking'],
            "devices" : ['binary_sensor.motion_floor', 'binary_sensor.motion_mirror'],
            }, ... }
        :return:
        """

        self.gen_hashmaps(dataset)

        if state_list is not None:
            self._state_list = dataset.encode
            # todo where is this used and what is state list set to
            raise ValueError
        else:
            self._state_list = dataset.get_state_list()

        self._observation_list = dataset.get_obs_list()
        self._dataset_repr = dataset.get_data_repr()

        self._dataset_conf = {
            'freq': dataset.get_freq(),
            'repr': dataset.get_data_repr(),
            'obs_type': 'bernoulli'
        }

        self._model_init(dataset)

    def gen_hashmaps(self, dataset):
        self._obs_lbl_hashmap = dataset.get_obs_lbl_hashmap()
        self._obs_lbl_rev_hashmap = dataset.get_obs_lbl_reverse_hashmap()
        self._state_lbl_hashmap = dataset.get_state_lbl_hashmap()
        self._state_lbl_rev_hashmap = dataset.get_state_lbl_reverse_hashmap()
        self.K = len(self._state_lbl_hashmap)
        self.D = len(self._obs_lbl_rev_hashmap)

    def are_hashmaps_created(self):
        return self._obs_lbl_hashmap is None \
               or self._obs_lbl_rev_hashmap is None \
               or self._state_lbl_hashmap is None \
               or self._state_lbl_rev_hashmap is None


    def register_act_info(self, act_data):
        self._act_data = self._encode_act_data(act_data)

    def register_loc_info(self, loc_data):
       self._loc_data = self._encode_location_data(loc_data)

    def _encode_act_data(self, act_data):
        """
        encodes the device names and activity names into numbers that the models
        can understand
        :param loc_data:
            is a list of loc data
            example:
            [ { "name" : "loc1", "activities" : ['cooking'],
            "devices" : ['binary_sensor.motion_floor', 'binary_sensor.motion_mirror'],
            },  ... ]
        :return:
            a list of location data, with the same structure correct encoded labels
             [ { "name" : "loc1",
                "activities" : [ 1 ],
                "devices" : [3, 7],
            },  ... ]
        """
        for activity in act_data:
            activity["name"] = self.encode_state_lbl(activity["name"])
        return act_data


    def _encode_location_data(self, loc_data):
        """
        encodes the device names and activity names into numbers that the models
        can understand
        :param loc_data:
            is a list of loc data
            example:
            [ { "name" : "loc1", "activities" : ['cooking'],
            "devices" : ['binary_sensor.motion_floor', 'binary_sensor.motion_mirror'],
            },  ... ]
        :return:
            a list of location data, with the same structure correct encoded labels
             [ { "name" : "loc1",
                "activities" : [ 1 ],
                "devices" : [3, 7],
            },  ... ]
        """
        for location in loc_data:
            new_act_list = []
            for activity in location['activities']:
                # todo the use of str vs. int is a hint that there is an
                # inconsistent use in the encoding and decoding in state labels
                new_act_list.append(str(self.encode_state_lbl(activity)))
            location['activities'] = new_act_list

            new_dev_list = []
            for device in location['devices']:
                new_dev_list.append(self.encode_obs_lbl(device))
            location['devices'] = new_dev_list
        return loc_data

    def _model_init(self, dataset, location_data):
        """
        this method has to be overriden by child classes
        :return:
        """
        raise NotImplementedError

    def predict_state_sequence(self, test_y):
        """
        for a given sequence the model should predict the underlying activities
        Parameters
        ----------
        test_y  array_like (... K)

        Returns
        -------
        pred_x  array_like (... K)
            the predicted activities
        """
        raise NotImplementedError

    def predict_obs_sequence(self, test_y):
        """
        for a given sequence the model should predict the next probable observation y_{t+1}
        at y_t
        Parameters
        ----------
        test_y  array_like (... K)

        Returns
        -------
        pred_y  array_like (... K)
        """
        raise NotImplementedError

    def train(self, dataset):
        """
         gets a dataset and trains the model on the data
         Important !!!!
         during the training the hashmaps of the model have to be filled up
        :param dataset:
        :return:
            the loss per iteration list
        """

        y_train = dataset.get_train_data()
        test_z, test_x = dataset.get_all_labeled_data() # states, obs
        lls = self._train(dataset)
        return lls


    def _train(self, dataset):
        raise NotImplementedError

    def get_state(self, seq):
        """
        returns the state the model is in given an observation sequence
        :return:
        """
        raise NotImplementedError

    def save_visualization(self, path_to_file):
        """
        save a visualization of the model to the given filepath
        :param path_to_file:
        :return:
        """
        raise NotImplementedError

    def draw(self):
        """
         somehow visualize the model
        :return: an image png or jpg
        """
        raise NotImplementedError

    def classify_multi(self, obs_seq, as_dict=True):
        """
        gets an observation sequence (at most times this is a window) and returns
        the most likely states
        :param obs_seq:
        :return:
        np array or dict
            with the most likely state
        """
        # encode obs_seq
        #obs_seq = self.obs_lbl_seq2enc_obs_seq(obs_seq)
        # array full of tuples with state_label and corresp. score
        import numpy as np
        scores = self._classify_multi(obs_seq)
        assert isinstance(scores, np.ndarray) and len(scores.shape) == 1

        if not as_dict:
            return scores
        else:
            act_score_dict = {}
            # decode state seq
            for i, score in enumerate(scores):
                label = self._state_lbl_rev_hashmap[i]
                act_score_dict[label] = score
            return act_score_dict

    def classify(self, obs_seq):
        """
        gets an observation sequence (at most times this is a window) and returns
        the most likely states
        Parameters
        ----------
        obs_seq np.ndarray
            list of np arrays containing the observations vectors
            e.g [ [0,1,0,0,1,...,0] , [0,0,1,0,...,1], ... ]

        Returns
        -------
        best_state (string)
            the most probable state the sequence is in

        """
        # encode obs_seq
        #obs_seq = self.obs_lbl_seq2enc_obs_seq(obs_seq)
        pred_state = self._classify(obs_seq)

        # decode state seq
        res = self._state_lbl_rev_hashmap[pred_state]
        return res

    def sample(self, n, obs_seq=None):
        """ samples
        Parameters
        ----------
        obs_seq
            optional observations which are used to sample stuff

        n   integer
            the amount of timesteps to generate

        Returns
        -------
        dec_states nd.array of strings

        dec_obs nd.array of strings
        """
        enc_states, enc_obs = self._sample(n, obs_seq)
        dec_states = self.decode_state_lbl_list(enc_states)
        dec_obs = self.decode_obs_lbl_list(enc_obs)
        return dec_states, dec_obs

    def decode_obs_lbl_list(self, lst):
        #return self._map_nparr_to_lbls(self._obs_lbl_rev_hashmap, lst)
        pass

    def decode_state_lbl_list(self, lst):
        return self._map_nparr_to_lbls(self._state_lbl_rev_hashmap, lst)


    def _map_nparr_to_lbls(self, dict, arr):
        """
        gets a encoded list and returns a list of labels
        Parameters
        ----------
        lst     list or np.ndarray
        Returns
        -------
            np.ndarray
        """
        N = len(arr)
        import numpy as np
        res = np.zeros((N), dtype=object)
        for i in range(N):
            val = arr[i]
            dec_val = dict[val]
            res[i] = dec_val
        return res


    def _sample(self, n, obs_seq=None):
        raise NotImplementedError

    def predict_next_obs(self, obs_seq):
        """
        predict the most probable device to change its state
        :param args:
        :return:
            label (string)
            the label of the device to change its state
        """
        #obs_seq = self.obs_lbl_seq2enc_obs_seq(obs_seq)
        last_obs = obs_seq[-1:][0]
        idx_next_obs = self._predict_next_obs(obs_seq)
        label = self._obs_lbl_rev_hashmap[idx_next_obs]
        return label

    def predict_prob_xnp1(self, obs_seq):
        """
        computes the probabilities of all observations to be the next
        :param obs_seq:
        :return:
            dictionary like: "{ 'sensor_name' : { 0 : 0.123, 1: 0.789 } , ... }"
        """
        #obs_seq = self.obs_lbl_seq2enc_obs_seq(obs_seq)
        arr = self._predict_prob_xnp1(obs_seq)
        res_dict = copy.deepcopy(self._obs_lbl_hashmap)
        for i, prob in enumerate(arr):
            label = self._obs_lbl_rev_hashmap[i]
            res_dict[label] = prob
        return res_dict

    def _classify(self, obs_seq):
        raise NotImplementedError

    def _classify_multi(self, obs_seq):
        raise NotImplementedError

    def _predict_next_obs(self, obs_seq):
        """
        has to return an array containing all the probabilities of the observations
        to be next
        :param obs_seq:
        :return:
        """
        raise NotImplementedError

    def _predict_prob_xnp1(self, obs_seq):
        raise NotImplementedError



from hbhmm.hmm._hmm_base import HMM
from hbhmm.hmm.hmm_log import HMM_log
from hbhmm.hmm.hmm_log_scaled import HMM_log_scaled
from hbhmm.hmm.hmm_scaled import HMM_Scaled
from hbhmm.hmm.distributions import ProbabilityMassFunction
from hassbrain_algorithm.models._model import Model
from hassbrain_algorithm.datasets.kasteren.kasteren import DatasetKasteren
import numpy as np

from hbhmm.hmm.probs import Probs


class _ModelHMM(Model):

    def __init__(self, controller):
        #self._cm = controller # type: Controller
        self._hmm = None # type: HMM

        # training parameters
        self._training_steps = 200
        self._epsilon = None
        self._use_q_fct = False

        Model.__init__(self, "test", controller)

    def __str__(self):
        if self._hmm is None:
            return "hmm has to be inits"
        else:
            return self._hmm.__str__()

    def use_q_fct(self, value):
        if value == True or value == False:
            self._use_q_fct = value
        else:
            raise ValueError

    def _model_init(self, dataset):
        K = len(self._state_list)
        D = len(self._observation_list)
        # init markov model in normal way
        self._hmm = HMM(self._state_list,
                        self._observation_list,
                        ProbabilityMassFunction,
                        initial_dist=None)
        init_pi = HMM.gen_rand_pi(K)
        self._hmm.set_pi(init_pi)

        em_matrix = HMM.gen_rand_emissions(K, D)
        self._hmm.set_emission_matrix(em_matrix)

        trans_mat = HMM.gen_rand_transitions(K)
        self._hmm.set_transition_matrix(trans_mat)


    def save_visualization(self, path_to_file):
        """
        save a visualization of the model to the given filepath
        :param path_to_file:
        :return:
        """
        graphviz_dot = self._hmm.generate_graphviz_dot_ext_lbl(
            self.save_visualization_helper_decode_labels
        )
        # strip png
        if path_to_file[-4:] == '.png':
            path_to_file = path_to_file[:-4]
            graphviz_dot.format = 'png'

        elif path_to_file[-4:] == '.jpg':
            path_to_file = path_to_file[:-4]
            graphviz_dot.format = 'jpg'
        else:
            graphviz_dot.format = 'png'

        graphviz_dot.render(filename=path_to_file)

    def save_visualization_helper_decode_labels(self, label):
        """
        :return:
        """
        return self.decode_state_lbl(label)


    def get_train_loss_plot_y_label(self):
        if self._use_q_fct:
            return 'Q(Theta, Theta_old)'
        else:
            return 'P(X|Theta)'


    def create_pred_act_seqs(self, dataset):
        state_lists, obs_lists = dataset.get_test_labels_and_seq()
        y_true = []
        y_pred = []
        for state_list, obs_list in zip(state_lists, obs_lists):
            tmp_true, tmp_pred = self._hmm.create_pred_act_seqs(state_list, obs_list)
            y_true.extend(tmp_true)
            y_pred.extend(tmp_pred)

        return y_true, y_pred

    def draw(self, act_retrieval_meth):
        return self._hmm.generate_graphviz_dot_ext_lbl(act_retrieval_meth)
        #vg.render('test.gv', view=True)

    def _train_loss_callback(self, hmm, loss, *args):
        # todo in log models convert to normal Probability
        """
        :param hmm:
        :param loss:
        :param args:
        :return:
        """
        # this is due to the the loss param is actually the likelihood of the P(X|Model)
        # therefore the loss can be 1 - P(X|Model)
        loss = 1-loss
        self._bench.train_loss_callback(hmm, loss)


    def train(self, dataset, args):
        if args is None:
            use_q_fct = False
        else:
            use_q_fct = args[0]

        self.use_q_fct(use_q_fct)

        if dataset.is_multi_seq_train():
            print('went here')
            obs_seq = dataset.get_train_seqs()
            self._hmm.train_seqs(
                set=obs_seq,
                epsilon=self._epsilon,
                steps=self._training_steps,
                callbacks=self._callbacks
            )

        else:
            obs_seq = dataset.get_train_seq()
            self._hmm.train(
                seq=obs_seq,
                epsilon=self._epsilon,
                steps=self._training_steps,
                q_fct=self._use_q_fct,
                callbacks=self._callbacks
            )

    def _classify(self, obs_seq):
        state_seq = self._hmm.viterbi(obs_seq)
        pred_state = state_seq[len(state_seq)-1]
        return pred_state

    def _classify_multi(self, obs_seq):
        """
        computes the last omega slice of viterbi which is
        equivalent to
        :param obs_seq:
        :return:
        """
        omega = self._hmm.viterbi_latt(obs_seq)
        N = len(obs_seq) - 1
        K = len(self._hmm._z)
        #print('*'*10)
        #print(omega)
        #print('*'*10)
        last_omega_slice = omega[N]
        # todo remove line below due to corrections
        norm = last_omega_slice.sum()
        last_omega_slice = last_omega_slice/norm
        last_omega_slice = Probs.np_prob_arr2exp(last_omega_slice)
        res = np.zeros((K), dtype=object)
        for i in range(K):
            res[i] = (self._hmm._z[i], last_omega_slice[i])
        return res


    def _predict_next_obs(self, obs_seq):
        next_obs = self._hmm.sample_observations(obs_seq, 1)
        #print('*'*100)
        #print('pre_next_obs', next_obs)
        next_obs = next_obs[0]
        #print('next_obs', next_obs)
        #print('*'*100)
        #next_obs = self._hmm.sample_observations(obs_seq, 1)[0]
        return next_obs

    def _predict_prob_xnp1(self, obs_seq):
        return self._hmm.predict_norm_probs_xnp(obs_seq)

class _ModelHMM_scaled(_ModelHMM):
    def __init__(self, controller):
        _ModelHMM.__init__(self, controller)

    def _model_init(self, dataset):
        state_list = dataset.get_state_list()
        observation_list = dataset.get_obs_list()
        K = len(state_list)
        D = len(observation_list)
        # init markov model in normal way
        #self._hmm = HMM_Scaled(state_list,
        self._hmm = HMM_Scaled(state_list,
                            observation_list,
                            ProbabilityMassFunction,
                            initial_dist=None)
        init_pi = HMM_Scaled.gen_rand_pi(K)
        self._hmm.set_pi(init_pi)

        em_matrix = HMM_Scaled.gen_rand_emissions(K,D)
        self._hmm.set_emission_matrix(em_matrix)

        trans_mat = HMM_Scaled.gen_rand_transitions(K)
        self._hmm.set_transition_matrix(trans_mat)


class ModelHMM_log(_ModelHMM):
    def __init__(self, controller):
        _ModelHMM.__init__(self, controller)

    def _model_init(self, dataset):
        state_list = dataset.get_state_list()
        observation_list = dataset.get_obs_list()
        K = len(state_list)
        D = len(observation_list)
        # init markov model in normal way
        #self._hmm = HMM_Scaled(state_list,
        self._hmm = HMM_log(state_list,
                                      observation_list,
                                      ProbabilityMassFunction,
                                      initial_dist=None)
        init_pi = HMM_log.gen_rand_pi(K)
        self._hmm.set_pi(init_pi)

        em_matrix = HMM_log.gen_rand_emissions(K,D)
        self._hmm.set_emission_matrix(em_matrix)

        trans_mat = HMM_log.gen_rand_transitions(K)
        self._hmm.set_transition_matrix(trans_mat)

    def get_train_loss_plot_y_label(self):
        if self._use_q_fct:
            return 'Q(Theta, Theta_old)'
        else:
            return 'log(P(X|Theta))'


    def _train_loss_callback(self, hmm, loss, *args):
        # todo in log models convert to normal Probability
        """
        :param hmm:
        :param loss:
        :param args:
        :return:
        """
        # this is due to the the loss param is actually the likelihood of the P(X|Model)
        # therefore the loss can be 1 - P(X|Model)
        #print('loss_before: ', loss)
        #print('Probs(1.0): ', Probs(1))
        #loss = Probs(1)/loss
        #print('1-loss: ', loss)
        #loss = loss.prob_to_norm()
        #print('exp(loss: ', loss)
        self._bench.train_loss_callback(hmm, float(loss))


    def _classify_multi(self, obs_seq):
        """
        computes the last omega slice of viterbi which is
        equivalent to
        :param obs_seq:
        :return:
        """
        omega = self._hmm.viterbi_latt(obs_seq)
        N = len(obs_seq) - 1
        K = len(self._hmm._z)
        print('lulululu'*50)
        print('omega: ', omega)
        last_omega_slice = omega[N]
        print('prob(x): ', last_omega_slice.sum())
        last_omega_slice = last_omega_slice/last_omega_slice.sum()
        print('omega_after: ', last_omega_slice)
        print('sum to 0: ', last_omega_slice.sum())
        last_omega_slice = Probs.np_prob_arr2exp(last_omega_slice)
        print('exp(omega_after: ', last_omega_slice)
        print('exp(sum to 1: ', last_omega_slice.sum())
        res = np.zeros((K), dtype=object)
        for i in range(K):
            res[i] = (self._hmm._z[i], last_omega_slice[i])
        return res

    def _predict_prob_xnp1(self, obs_seq):
        res = self._hmm.predict_probs_xnp(obs_seq)
        return Probs.np_prob_arr2exp(res)

class ModelHMM_log_scaled(ModelHMM_log):
    def __init__(self, controller):
        _ModelHMM.__init__(self, controller)
        self._training_steps = 10

    def _model_init(self, dataset):
        state_list = dataset.get_state_list()
        observation_list = dataset.get_obs_list()
        K = len(state_list)
        D = len(observation_list)
        self._hmm = HMM_log_scaled(state_list,
                            observation_list,
                            ProbabilityMassFunction,
                            initial_dist=None)
        init_pi = HMM_log.gen_rand_pi(K)
        self._hmm.set_pi(init_pi)

        em_matrix = HMM_log.gen_rand_emissions(K,D)
        self._hmm.set_emission_matrix(em_matrix)

        trans_mat = HMM_log.gen_rand_transitions(K)
        self._hmm.set_transition_matrix(trans_mat)

# test to use alpha forward recursion instead of viterbi
class HMMForward(ModelHMM_log_scaled):
    def __init__(self, controller):
        ModelHMM_log_scaled.__init__(self, controller)

    def _classify(self, obs_seq):
        norm_alpha, cn = self._hmm.forward(obs_seq)
        #print('went here1')
        #print('normalized_alpha: ', norm_alpha)
        alpha = self._hmm.nalpha_to_alpha(norm_alpha, cn)
        #print('went here2')
        #print('alpha: ', alpha)
        #print('alpha: ', Probs.np_prob_arr2exp(alpha))
        #print('last_alpha: ', Probs.np_prob_arr2exp(alpha[-1:]))
        last_alpha_vals = Probs.np_prob_arr2exp(alpha[-1:])
        idx = np.argmax(last_alpha_vals)
        #print('idx: ', idx)
        pred_state = self._hmm._z[idx]
        #print('act: ', pred_state)

        return pred_state


    def _classify_multi(self, obs_seq):
        """
        computes the last slice of alpha forward pass which is
        the probability of being in state z at time n
        equivalent to
        :param obs_seq:
        :return:
        """
        norm_alpha, cn = self._hmm.forward(obs_seq)
        #print('went here1')
        #print('normalized_alpha: ', norm_alpha)
        alpha = self._hmm.nalpha_to_alpha(norm_alpha, cn)
        last_omega_slice = Probs.np_prob_arr2exp(alpha[-1:][0])
        last_omega_slice = last_omega_slice/last_omega_slice.sum()
        last_omega_slice = Probs.np_prob_arr2exp(last_omega_slice)
        print('lululu'*100)
        K = len(self._hmm._z)
        res = np.zeros((K), dtype=object)
        for i in range(K):
            res[i] = (self._hmm._z[i], last_omega_slice[i])
        return res



from ray import tune



# Random Forest sktime
clf_random_forest = dict(
    clf__n_estimators=tune.qrandint(10, 400),
    clf__warm_start=tune.choice([True, False]),
    clf__max_depth=tune.qrandint(1, 200)
)


clf_HIVECOTEV2 = dict(
    stc_params=dict(
        n_shapelet_samples=10000, # Default = 10000
    ),
    drcif_params=dict(),
    arsenal_params=dict(),
    tde_params=dict(),
)


# https://www.sktime.net/en/latest/api_reference/auto_generated/sktime.classification.dictionary_based.BOSSEnsemble.html
clf_BOSSEnsemble = dict(
    clf_boss__max_ensemble_size=tune.randint(300, 700),    # default=500
    clf_boss__max_win_len_prop=tune.uniform(0.5, 1.5),     # default=1
    clf_boss__min_window=tune.randint(10, 50),             # default=10
    clf_boss__alphabet_size=tune.randint(2,5),             # default=2

)


clf_ROCKET = dict(
    clf_rocket__num_kernels=tune.randint(5000, 15000),       # default=10000
    clf_rocket__rocket_transform=tune.choice(['rocket', 'minirocket', 'multirocket']),
    clf_rocket__n_features_per_kernel=tune.randint(10, 50),  # default=32
    clf_rocket__use_multivariate='yes'
)


clf_CNN = dict(
    clf_cnn__n_epochs=100,                          # default=2000
    clf_cnn__batch_size=tune.randint(1, 32),        # default=16
    clf_cnn__kernel_size=tune.randint(3, 16),       # default=7
    clf_cnn__avg_pool_size=tune.randint(3, 5),      # default=3
    clf_cnn__use_bias=tune.choice([True, False])    # default=true

)


from hassbrain_algorithm.datasets._dataset import _Dataset
from hassbrain_algorithm.models._model import Model
import pandas as pd
import numpy as np

def _str2boolean(arr):
    return arr == 'True'

def _boolean2str(arr):
    res = arr.astype(str)
    return res

def _num2bool(arr):
    return arr.astype(bool)

def _onehot(train_z, num_classes):
    T = len(train_z)
    res = np.zeros((T, num_classes), dtype=np.float64)
    for t in range(T):
        res[t][train_z[t]] = 1.
    return res

def compute_explanation(model : Model, x, z):
    """ generates an explanation how the model came to its result
    Parameters
    ----------
    model : Model
        a hassbrain model
    x : np.ndarray
        the raw array conatining the sensor values
    Returns
    -------

    """
    from matplotlib.pyplot import Figure
    wrapped_model = ModelWrapper(model)
    class_names = model.get_state_lbl_lst()
    feature_names = model.get_obs_lbl_lst()


    cat_idxs = [i for i in range(len(class_names))]
    categorical_names = {}
    for i in cat_idxs:
        categorical_names[i] = {}
        categorical_names[i][0] = "off"
        categorical_names[i][1] = "on"

    from skater.core.local_interpretation.lime.lime_tabular import LimeTabularExplainer
    exp = LimeTabularExplainer(
        x,
        mode='classification',
        training_labels=z,
        feature_names=feature_names,
        categorical_features=cat_idxs,
        categorical_names=categorical_names,
        class_names=class_names)

    fig = exp.explain_instance(x[0], wrapped_model.predict_proba).as_pyplot_figure()


def compute_feature_importance(model : Model, X : np.ndarray, z : np.ndarray):
    """ calculates the feature importance, the impact on prediction on the dataset
    Parameters
    ----------
    model : Model
        a model of
    X : array-like
        the array the importance should be calculated on
    z : array-like
        the corresponding labels
    Returns
    -------
    res : pd.Dataframe (1, D)

    """
    from skater.model import InMemoryModel
    from skater.core.explanations import Interpretation
    from matplotlib.pyplot import Figure
    wrapped_model = ModelWrapper(model)
    class_names = model.get_state_lbl_lst()
    feature_names = model.get_obs_lbl_lst()

    # this has to be done in order for skater to recognize the values as categorical and not numerical
    X = _boolean2str(X)

    # create interpretation
    interpreter = Interpretation(X,
                                 #class_names=class_names,
                                 feature_names=feature_names)

    # create model
    # supports classifiers with or without probability scores
    examples = X[:10]
    skater_model = InMemoryModel(wrapped_model.predict,
                                #target_names=class_names,
                                feature_names=feature_names,
                                model_type='classifier',
                                unique_values=class_names,
                                probability=False,
                                examples=examples)

    # only do this for cross_entropy
    #train_z = onehot(train_z, model.K)
    interpreter.load_data(X,
                          training_labels=z,
                          feature_names=feature_names)

    # todo flag for deletion (3lines below)
    #    if this can savely be deleted
    tmp = interpreter.data_set.feature_info
    for key, val in tmp.items():
        val['numeric'] = False
    fig, axes = interpreter.feature_importance.save_plot_feature_importance(
         skater_model,
         ascending=True,
         ax=None,
         progressbar=False,
         # model-scoring: difference in log_loss or MAE of training_labels
         # given perturbations. Note this vary rarely makes any significant
         # differences
         method='model-scoring')
        # corss entropy or f1 ('f1', 'cross_entropy')
         #scorer_type='cross_entropy') # type: Figure, axes
         #scorer_type='f1') # type: Figure, axes

        # cross_entropy yields zero

    fig.show()


class ModelWrapper():
    def __init__(self, model : Model):
        self.model = model

    def predict_old(self, X):
        """
        Parameters
        ----------
        X

        Returns
        -------
        z_labeld_pred : array-like
            an array of the indicies of the class labels
        """

        if X.dtype == np.float:
            X = _num2bool(X)
        elif X.dtype != bool:
            X = _str2boolean(X)

        z_pred = self.model._hmm.most_likely_states(X)
        z_labeld_pred = np.zeros_like(z_pred, dtype=object)
        for i, item in enumerate(z_pred):
            z_labeld_pred[i] = str('z' + str(z_pred[i]))
        return z_labeld_pred

    def predict(self, X):
        """
        Parameters
        ----------
        X

        Returns
        -------
        z_labeld_pred : array-like
            an array of the indicies of the class labels
        """

        if X.dtype == np.float:
            X = _num2bool(X)
        elif X.dtype != bool:
            X = _str2boolean(X)

        # todo change this IMPORTANT
        try:
            z_pred = self.model._hmm.most_likely_states(X)
        except:
            z_pred = self.model._hsmm.most_likely_states(X)

        z_labeld_pred = np.zeros_like(z_pred, dtype=object)
        for i, item in enumerate(z_pred):
            lbl = self.model.decode_state_lbl(item)
            z_labeld_pred[i] = lbl
        return z_labeld_pred

    def predict_proba(self, X):
        """Predict class probabilities for X as an array of probabilities.
        This is used for cross entropy comparision

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            The input samples. Internally, it will be converted to
            ``dtype=np.float32`` and if a sparse matrix is provided
            to a sparse ``csr_matrix``.

        Raises
        ------
        AttributeError
            If the ``loss`` does not support probabilities.

        Returns
        -------
        p : array, shape (n_samples, n_classes)
            The class probabilities of the input samples. The order of the
            classes corresponds to that in the attribute `classes_`.
        """
        if len(X.shape) == 1:
            X = np.reshape(X, (-1, len(X)))
        T, _ = X.shape
        K = self.model.K

        if X.dtype == np.float:
            X = _num2bool(X)

        resnp = np.zeros((T, K), dtype=np.float64)
        for i, row in enumerate(X):
            row = np.reshape(row, (-1, len(row)))
            tmp = self.model.classify_multi(row)
            resnp[i, :] = tmp
        return resnp



from copy import deepcopy

import dash
import dash.dcc as dcc
import dash.html as html
import dash_bootstrap_components as dbc
import pandas as pd
from dash.dependencies import *  # TODO remove if nothing else is imported
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate
from dask.delayed import delayed
import numpy as np
from pyadlml.constants import (ACTIVITY, DEVICE, END_TIME, START_TIME, TIME,
                               VALUE)
from pyadlml.dataset._core.acts_and_devs import label_data
from pyadlml.dataset._core.activities import add_other_activity, create_empty_activity_df
from pyadlml.dataset._datasets.aras import _create_device_df
from pyadlml.dataset.plot.plotly.util import dash_get_trigger_element, dash_get_trigger_value
from pyadlml.dataset.plot.plotly.activities import *
from pyadlml.dataset.plot.plotly.acts_and_devs import *
from pyadlml.dataset.plot.plotly.dashboard.callbacks import (
    _create_activity_tab_callback, _create_acts_vs_devs_tab_callback, _initialize_activity_toggle_cbs)
from pyadlml.dataset.plot.plotly.dashboard.layout import (
    BIN_SIZE_SLIDER, DEV_DENS_SLIDER, LAG_SLIDER, activities_layout, acts_n_devs_layout,
    acts_vs_devs_layout, device_layout_graph_bottom, devices_layout)
from pyadlml.dataset.plot.plotly.devices import bar_count as dev_bar_count
from pyadlml.dataset.plot.plotly.devices import boxplot_state
from pyadlml.dataset.plot.plotly.devices import device_iei as dev_iei
from pyadlml.dataset.plot.plotly.devices import event_density
from pyadlml.dataset._core.activities import ActivityDict
from pyadlml.dataset.stats.acts_and_devs import (contingency_table_events,
                                                 contingency_table_states)
from pyadlml.dataset.util import (activity_order_by, device_order_by, num_to_timestamp,
                                  select_timespan, timestamp_to_num)
from pyadlml.dataset.plot.plotly.devices import plotly_event_correlogram as dev_cc
from pyadlml.dataset.plot.plotly.util import deserialize_range, serialize_range, range_from_fig
from pyadlml.constants import ts2str, str2ts

def _get_plot_height_devs(nr_devs):
    # Determine the plot height and fontsize for activity plots
    if nr_devs < 20:
        return 350
    elif nr_devs < 30:
        return 400
    elif nr_devs < 50:
        return 500
    elif nr_devs < 70:
        return 700
    else:
        return 800


def _get_plot_height_acts(nr_acts):
    if nr_acts < 20:
        return 350
    elif nr_acts < 25:
        return 450
    else:
        return 350


def dashboard(app, name, embedded=False, df_acts=None, df_devs=None, dev2area=None, start_time=None, end_time=None):
    """ Creates a dashboard

    Note
    -----
    All parameter may be 'None' if the app is populated during runtime

    """

    df_acts = ActivityDict.wrap(df_acts)

    # Since performance rendering issues arise with numbers greater than 40000
    # device datapoints make a preselection
    max_points = 40000
    if df_devs is not None and len(df_devs) > max_points:
        set_et = df_devs.iat[max_points, df_devs.columns.get_loc(TIME)]
        curr_df_devs, curr_df_acts = select_timespan(
            df_acts=df_acts, df_devs=df_devs, start_time=start_time,
            end_time=set_et, clip_activities=True
        )
    else:
        curr_df_devs = df_devs
        curr_df_acts = df_acts
        set_et = end_time
    
    nr_devs = len(df_devs[DEVICE].unique()) if df_devs is not None else 0
    nr_acts = df_acts.nr_acts() if df_acts is not None else 0

    if start_time is None:
        start_time = min(df_devs[TIME].iloc[0], df_acts.min_starttime())
        start_time = start_time.floor('D')
    if end_time is None:
        end_time = max(df_devs[TIME].iloc[-1], df_acts.max_endtime())
        end_time = end_time.ceil('D')

    # TODO refactor: ugly as hell defined in 76
    if df_devs is None or len(df_devs) <= max_points:
        set_et = end_time

    plot_height_devs = _get_plot_height_devs(nr_devs)
    plot_height_acts = _get_plot_height_acts(nr_acts)
    is_top_down_view = (nr_devs < 60)
    plot_height_ands = (plot_height_devs+50 if is_top_down_view else 1300)

    # Get Layout
    layout_activities = [activities_layout(df, k, plot_height_acts) 
                         for k, df in curr_df_acts.items()]
    layout_devices = devices_layout(curr_df_devs, True, plot_height_devs)
    layout_acts_vs_devs = [acts_vs_devs_layout(df, k, df_devs, plot_height=plot_height_acts) 
                           for k, df in curr_df_acts.items()]

    # Generate tabs for multiple persons
    activity_tabs = [dbc.Tab(l, label=f'Act: {act_name}', tab_id=f'tab-acts-{act_name}') 
                     for act_name, l in zip(curr_df_acts.keys(), layout_activities)]
    acts_vs_devs_tabs = [
                 dbc.Tab(l, label=f'Act: {act_name} ~ Devices', tab_id=f'tab-acts_vs_devs-{act_name}')
                 for act_name, l in zip(curr_df_acts.keys(), layout_acts_vs_devs)]
    
    def layout_content(and_layout, tabs_and_content):
        if is_top_down_view:
            return [
                and_layout, 
                html.Br(),
                html.Div([
                    *tabs_and_content
                ]),
            ]
        else:
            return [
                dbc.Row([
                    dbc.Col([
                        and_layout, 
                    ], md=7),
                    dbc.Col([
                        html.Div([
                            *tabs_and_content
                        ]),
                    ], md=5),
                ], style=dict(width="2800px", height="1500px"))
            ]

    content = [
        dcc.Input(id='act_assist_path', type='hidden', value='filler text'),
        dcc.Input(id='subject_names', type='hidden', value='filler text'),
        html.H2(children=f'Dashboard: {name}'),

        *layout_content(
            acts_n_devs_layout(df_devs, df_acts, start_time, end_time, set_et, plot_height_ands),
            [dbc.Tabs(
                [*activity_tabs,
                    dbc.Tab(layout_devices, label='Devices', tab_id='tab-devs'),
                    *acts_vs_devs_tabs
                    ], id='tabs', active_tab=f'tab-acts-{list(curr_df_acts.keys())[0]}',
            ),
            html.Div(id="content"),]
        )
    ]

    if embedded:
        layout = dbc.Container(
            children=content, 
            style={'width': 1300, 'margin': 'auto'}
        )
    else:
        width = 1320 if is_top_down_view else 2200
        layout = dbc.Container(
            children=content,
            style={'width': width}
        )
    app.layout = layout

    df_acts.apply(add_other_activity)

    create_callbacks(app, df_acts, df_devs, start_time, end_time, dev2area, 
                     plot_height_acts, plot_height_devs, plot_height_ands
    )

def _sel_avd_event_click(avd_event_select, curr_df_acts, curr_df_devs):
    """Select devices and activities from click data of the event contingency table"""

    # Get relevant acts and devs
    activity = avd_event_select['points'][0]['y']
    device = avd_event_select['points'][0]['x']
    df_devs = curr_df_devs[curr_df_devs[DEVICE] == device].copy()
    df_acts = curr_df_acts[curr_df_acts[ACTIVITY] == activity].copy()

    # Select devices
    sel_devices = label_data(df_devs, df_acts).dropna()
    sel_devices = sel_devices[[TIME, DEVICE, VALUE]]

    # Select activities
    sel_activities = create_empty_activity_df()
    for _, sel_dev in sel_devices.iterrows():
        mask = (df_acts[START_TIME] < sel_dev[TIME]) \
               & (sel_dev[TIME] < df_acts[END_TIME])
        sel_activities = pd.concat([sel_activities, df_acts[mask]])
    sel_activities = sel_activities.drop_duplicates()
    return sel_devices, sel_activities


def _sel_point_to_activities(selection):
    """ Gets the selected activities from the boxplot or violin-plot and
        asdf
    """
    user_sel_activities = []
    points = selection['points']
    for usa in points:
        user_sel_activities.append([usa['customdata'][0], usa['customdata'][1], usa['y']])

    df_usa = pd.DataFrame(user_sel_activities, columns=[START_TIME, END_TIME, ACTIVITY])
    df_usa[START_TIME] = pd.to_datetime(df_usa[START_TIME])
    df_usa[END_TIME] = pd.to_datetime(df_usa[END_TIME])
    return df_usa

def _sel_act_bar(selection, df_acts):
    """ Retrieve the activity from a click on an activity bar in a
        bar plot e.g count, cummulative and filter the activity dataframe

    """
    dev = selection['points'][0]['y']
    return df_acts[df_acts[ACTIVITY] == dev]


def _sel_dev_bar(selection, df_devs):
    """ Retrieve the device from a click on the device bar the event count
        graphand filter the device dataframe

    """
    dev = selection['points'][0]['y']
    return df_devs[df_devs[DEVICE] == dev]

def _sel_dev_fraction(selection, df_devs):
    """ Retrieve the device from a click on the device bar the event count
        graphand filter the device dataframe

    """
    dev = selection['points'][0]['y']
    cat = selection['points'][0]['customdata'][1]
    mask = (df_devs[DEVICE] == dev) & (df_devs[VALUE] == cat)
    return df_devs[mask].copy()

def _sel_dev_bp_selection(dev_boxplot_select, df_devs):
    """ Gets point data from the state boxplot and creates
        an device-dataframe
    """
    points = dev_boxplot_select['points']
    sd = pd.DataFrame(columns=[TIME, END_TIME, DEVICE, VALUE])
    for p in points:
        duration = pd.Timedelta(p['customdata'][0])
        time = p['customdata'][1]
        dev = p['y']
        mask = (df_devs[DEVICE] == dev) & (df_devs[TIME] == time)
        df = df_devs[mask]
        df[END_TIME] = df[TIME] + duration
        sd = pd.concat([sd, df])

    return sd


def _sel_dev_iei(dev_iei_select, df_devs, scale, fig):

    per_device = not (len(fig['data']) == 1)
    df = df_devs.copy()
    df_res = pd.DataFrame(columns=[TIME, DEVICE, VALUE])
    df['diff'] = df[TIME].shift(-1) - df[TIME]
    size = fig['data'][0]['xbins']['size']

    for p in dev_iei_select['points']:
        left, right = p['x'] - size/2, p['x'] + size/2
        if scale == 'log':
            left, right = np.exp(left), np.exp(right)
        mask = (pd.to_timedelta(left, unit='s') <= df['diff']) \
               & (df['diff'] <= pd.to_timedelta(right, unit='s'))
        if per_device:
            mask = mask & (df[DEVICE] == p['customdata'])

        count = sum(mask)
        sel_count = len(p['pointNumbers'])
        assert count == sel_count, f'Lookup error: The selected device number {sel_count} != {count}'
        df_res = pd.concat([df_res, df_devs[mask]])

    return df_res


def _sel_dev_density(selection, df_devs, dt):

    from pyadlml.dataset.plot.plotly.dashboard.layout import DEV_DENS_SLIDER
    df = df_devs.copy().set_index(TIME, drop=False)
    dt = pd.Timedelta(DEV_DENS_SLIDER[dt])
    df_res = pd.DataFrame(columns=[TIME, DEVICE, VALUE])

    for p in selection['points']:
        dev = p['y']
        count = p['z']
        mid = pd.Timestamp(p['x'])
        tmp = df[df[DEVICE] == dev].between_time(
                    start_time=(mid - dt/2).time(),
                    end_time=(mid + dt/2).time())\
              .reset_index(drop=True)

        assert count == len(tmp.index)
        df_res = pd.concat([df_res, tmp])

    if df_res.empty:
        return None
    else:
        return df_res


def _sel_act_trans_sel(act_trans_select, df_acts):
    """ Gets point data from the state boxplot and creates
        an device-dataframe
    """
    df = df_acts.copy()\
        .sort_values(by=START_TIME)\
        .reset_index(drop=True)
    act1 = act_trans_select['points'][0]['x']
    act2 = act_trans_select['points'][0]['y']
    df[ACTIVITY + '2'] = df[ACTIVITY].shift(-1)
    idx = df[(df[ACTIVITY] == act1) & (df[ACTIVITY + '2'] == act2)].index.values
    df4 = df.iloc[[*idx, *(idx+1)]]
    df4 = df4[[START_TIME, END_TIME, ACTIVITY]]\
            .sort_values(by=START_TIME)\
            .reset_index(drop=True)
    return df4


def _initialize_toggle_callbacks(app):


    @app.callback(
        Output("clps-acts-n-devs", "is_open"),
        [Input("clps-acts-n-devs-button", "n_clicks")],
        [State("clps-acts-n-devs", "is_open")],
    )
    def toggle_collapse(n, is_open):
        return bool(n) ^ bool(is_open)

    @app.callback(
        Output("clps-dev-iei", "is_open"),
        [Input("clps-dev-iei-button", "n_clicks")],
        [State("clps-dev-iei", "is_open")],
    )
    def toggle_collapse_act_bp(n, is_open):
        return bool(n) ^ bool(is_open)

    @app.callback(
        Output("clps-dev-bar", "is_open"),
        [Input("clps-dev-bar-button", "n_clicks")],
        [State("clps-dev-bar", "is_open")],
    )
    def toggle_collapse_act_bp(n, is_open):
        return bool(n) ^ bool(is_open)

    @app.callback(
        Output("clps-dev-bp", "is_open"),
        [Input("clps-dev-bp-button", "n_clicks")],
        [State("clps-dev-bp", "is_open")],
    )
    def toggle_collapse_act_bp(n, is_open):
        return bool(n) ^ bool(is_open)

    @app.callback(
        Output("clps-dev-density", "is_open"),
        [Input("clps-dev-density-button", "n_clicks")],
        [State("clps-dev-density", "is_open")],
    )
    def toggle_collapse_act_bp(n, is_open):
        return bool(n) ^ bool(is_open)

    @app.callback(
        Output("clps-dev-cc", "is_open"),
        [Input("clps-dev-cc-button", "n_clicks")],
        [State("clps-dev-cc", "is_open")],
    )
    def toggle_collapse_act_bp(n, is_open):
        return bool(n) ^ bool(is_open)

def create_callbacks(app, dct_acts, df_devs, start_time, end_time, dev2area, plt_height_acts, plt_height_devs, plt_height_ands):

    def gen_trigger(id, val):
        return html.Div(id=id, style=dict(display="none"), **{"data-value": val})

    _initialize_toggle_callbacks(app)

    for key in dct_acts.keys():
        _initialize_activity_toggle_cbs(app, key)

    @app.callback(
        Output("and_clipboard", "content"),
        Input("graph-acts_n_devs", "clickData"),
        Input("and_btn_copy_range", "n_clicks"),
        State("and_dev-type", "value"),
        State("graph-acts_n_devs", "figure")
    )
    def element_to_clipboard(tmp, btn_cr, dev_type, fig_and):

        ctx = dash.callback_context
        trigger_val =  dash_get_trigger_value(ctx)
        trigger = dash_get_trigger_element(ctx)
        is_copy_range = (trigger == 'and_btn_copy_range')
        if trigger_val is None\
            or (tmp is None and not is_copy_range):
            raise PreventUpdate

        if not is_copy_range:
            dp = tmp['points'][0]
            is_activity = (dp['y'] == 'Activity')
            is_dev_state = not is_activity and dev_type == 'state'
            is_dev_event = not is_activity and dev_type == 'event'

        if is_copy_range:
            range = range_from_fig(fig_and)
            s = f'{ts2str(range[0])},{ts2str(range[1])}'
        elif is_activity:
            start_time = dp['base']
            end_time = dp['x']
            s = f'\'{start_time}\',\'{end_time}\',#activity'
        elif is_dev_state:
            start_time = dp['base']
            end_time = dp['customdata'][1]
            dev = dp['y']
            val = dp['customdata'][2]
            s = f'\'{start_time}\',\'{end_time}\',\'{dev}\',\'{val}\''
        elif is_dev_event:
            time = dp['x']
            dev = dp['y']
            val = dp['customdata']
            s = f'\'{time}\',\'{dev}\',\'{val}\''
        else:
            raise NotImplementedError
        return s

    act_outputs = []
    act_inputs = []
    act_states = []
    for subj_name  in dct_acts.keys():
        act_outputs.extend([
            Output(f'act-trigger-{subj_name}', 'children'),
            Output(f'act-curr-sel-{subj_name}', 'children'),
            Output(f'act-curr-sel-store-{subj_name}', 'data'),
            Output(f'avd-trigger-{subj_name}', 'children'),
        ])
        act_inputs.extend([
            Input(f'acts_graph-boxplot-{subj_name}', 'selectedData'),
            Input(f'acts_graph-transition-{subj_name}', 'clickData'),
            Input(f'acts_graph-bar-{subj_name}', 'clickData'),
            Input(f'avd_graph-event-contingency-{subj_name}', 'clickData'),
        ])
        act_states.extend([
            State(f'act-curr-sel-store-{subj_name}', 'data'), 
            State(f'act-curr-sel-{subj_name}', 'children'),
        ])

    @app.callback(
        output=[
            Output('graph-acts_n_devs', 'figure'),
            Output('and_act-order', 'data'),
            Output('and_dev-order', 'data'),
            Output('and_reset_sel', 'disabled'),
            Output('dev-trigger', 'children'),
            Output('dev-curr-sel', 'children'),
            Output('dev-curr-sel-store', 'data'),
            *act_outputs
        ],
        inputs=[
            # Input methods
            Input('range-slider', 'value'),
            Input('select-activities', 'value'),
            Input('select-devices', 'value'),
            Input('and_act-order-trigger', 'value'),
            Input('and_dev-order-trigger', 'value'),
            Input('and_dev-type', 'value'),

            # Selected devices
            Input('devs_graph-bar', 'clickData'),
            Input('devs_graph-boxplot', 'selectedData'),
            Input('devs_graph-iei', 'selectedData'),
            Input('devs_graph-fraction', 'clickData'),

            Input('devs_graph-density', 'clickData'),
            Input('and_reset_sel', 'n_clicks'),

            # Selected activities
            *act_inputs
        ],
        state=[
            State('devs_dens-slider', 'value'),
            State('devs_iei-scale', 'value'),
            State('devs_graph-iei', 'figure'),
            State('and_act-order', 'data'),
            State('and_dev-order', 'data'),
            State('dev-curr-sel', 'children'),
            State('dev-curr-sel-store', 'data'), 
            *act_states
        ]
    )
    def update_acts_n_devs(*args):

        nr_subjects = len(dct_acts.keys())

        idx = 12
        rng, sel_activities, sel_devices, act_order_trigger, dev_order_trigger, dev_type_trigger, \
        dev_bar_select, dev_boxplot_select, dev_iei_select, dev_fraction_select, \
        dev_density_select, reset_sel = args[:idx]    # [0:12] -> 1el, ..., 12 el

        for i, subj_name in enumerate(dct_acts.keys()):
            j = idx+i+i*nr_subjects
            k, l, m = j+1, j+2, j+3
            locals()[f'act_boxplot_select_{subj_name}'] = args[j]   # [13] -> 14 el  
            locals()[f'act_trans_select_{subj_name}'] = args[k]     # [14] -> 15 el
            locals()[f'act_bar_select_{subj_name}'] = args[l]       # [15] -> 16 el
            locals()[f'avd_event_select_{subj_name}'] = args[m]     # [16] -> 17 el

        idx = idx+4*nr_subjects # 16 - [17:25] -> 19el,...,25 el
        dev_density_dt, dev_iei_scale, dev_iei_fig, act_order, dev_order,  \
        dev_curr_sel, dev_curr_sel_store =  args[idx:idx+7]

        idx = idx+7
        for i, subj_name in enumerate(dct_acts.keys()):
            j, k = idx+i*nr_subjects, idx+i*nr_subjects+1
            locals()[f'act_curr_sel_store_{subj_name}'] = args[j]
            locals()[f'act_curr_sel_{subj_name}'] = args[k]


        ctx = dash.callback_context

        if dash_get_trigger_value(ctx) is None:
            raise PreventUpdate

        # Catch block if no act-order or dev-order is initialized
        try:
            act_order = json.loads(act_order)
        except TypeError:
            act_order = None
        try:
            dev_order = json.loads(dev_order)
        except TypeError:
            dev_order = None

        try:
            df_dev_curr_sel = pd.read_json(dev_curr_sel_store)
        except:
            df_dev_curr_sel = None

        origin_act_trans_selection = any([_is_trigger(ctx, f'acts_graph-transition-{sub}') for sub in dct_acts.subjects()])
        origin_act_bp_selection = any([_is_trigger(ctx, f'acts_graph-boxplot-{sub}') for sub in dct_acts.subjects()])
        origin_act_bar_selection = any([_is_trigger(ctx, f'acts_graph-bar-{sub}') for sub in dct_acts.subjects()])

        origin_and_event_selection = _is_trigger(ctx, 'avd_graph-event-contingency')
        origin_dev_bp_selection = _is_trigger(ctx, 'devs_graph-boxplot')
        origin_dev_iei_selection = _is_trigger(ctx, 'devs_graph-iei')
        origin_dev_density_selection = _is_trigger(ctx, 'devs_graph-density')
        origin_dev_bar_selection = _is_trigger(ctx, 'devs_graph-bar')
        origin_dev_fraction_selection = _is_trigger(ctx, 'devs_graph-fraction')

        data_selection_changed = _is_trigger(ctx, 'range-slider') or _is_trigger(ctx, 'select-devices') \
                                 or _is_trigger(ctx, 'select-activities')

        user_sel_activities = origin_act_trans_selection \
                              or origin_act_bp_selection \
                              or origin_and_event_selection \
                              or origin_act_bar_selection
        user_sel_devices = origin_and_event_selection \
                           or origin_dev_bp_selection \
                           or origin_dev_iei_selection \
                           or origin_dev_density_selection \
                           or origin_dev_bar_selection \
                           or origin_dev_fraction_selection

        reset_selection = _is_trigger(ctx, 'and_reset_sel') or data_selection_changed


        # Create start- and end-timestamps from range slider
        st = num_to_timestamp(rng[0], start_time=start_time, end_time=end_time)
        et = num_to_timestamp(rng[1], start_time=start_time, end_time=end_time)
        curr_df_devs, curr_dct_acts = select_timespan(df_acts=dct_acts, df_devs=df_devs,
                                                     start_time=st, end_time=et, clip_activities=True)



        # Filter selected devices and activities
        curr_df_devs = curr_df_devs[curr_df_devs[DEVICE].isin(sel_devices)]
        curr_dct_acts = deepcopy(curr_dct_acts)
        for key, df in dct_acts.items():
            curr_dct_acts[key] = df[df[ACTIVITY].isin(sel_activities)]

        # Filter for acts-n-devs plot
        df_devs1 = df_devs[df_devs[DEVICE].isin(sel_devices)]

        dct_acts1 = deepcopy(dct_acts)
        for key, df in dct_acts.items():
            dct_acts1[key] = df[df[ACTIVITY].isin(sel_activities)]

        try:
            df_act_curr_sel = ActivityDict.read_json(act_curr_sel_store)
        except:
            df_act_curr_sel = ActivityDict()

        devs_usel_state = None
        # Create activity and device order if needed
        if _is_trigger(ctx, 'and_dev-order-trigger') or dev_order is None:
            dev_order = device_order_by(curr_df_devs, dev_order_trigger, dev2area)
        if _is_trigger(ctx, 'and_act-order-trigger') or act_order is None:
            act_order = activity_order_by(curr_dct_acts, act_order_trigger)


        # Create dataframes with user selection
        for subj in dct_acts.subjects():
            if (_is_trigger(ctx, f'acts_graph-transition-{subj}') \
                or locals()[f'act_curr_sel_{subj}']  == f'acts_graph-transition-{subj}')\
                and not reset_selection:
                df_act_curr_sel[subj] = _sel_act_trans_sel(locals()[f'act_trans_select_{subj}'],
                                                    df_acts=curr_dct_acts[subj])
                locals()[f'act_curr_sel_{subj_name}'] = f'acts_graph-transition-{subj}'

            if (_is_trigger(ctx, f'acts_graph-boxplot-{subj}') \
                or locals()[f'act_curr_sel_{subj}'] == f'acts_graph-boxplot-{subj}')\
                and not reset_selection:
                df_act_curr_sel[subj] = _sel_point_to_activities(locals()[f'act_boxplot_select_{subj}'])
                locals()[f'act_curr_sel_{subj_name}'] = f'acts_graph-boxplot-{subj}'

            if (_is_trigger(ctx, f'acts_graph-bar-{subj}') \
                or locals()[f'act_curr_sel_{subj}'] == f'acts_graph-bar-{subj}')\
                and not reset_selection:
                df_act_curr_sel[subj] = _sel_act_bar(locals()[f'act_bar_select_{subj}'], df_acts=curr_dct_acts[subj])
                locals()[f'act_curr_sel_{subj_name}'] = f'acts_graph-bar-{subj}'

            if _is_trigger(ctx, f'avd_graph-event-contingency-{subj}')\
                and not reset_selection:
                df_dev_curr_sel, df_act_curr_sel[subj] = _sel_avd_event_click(
                    locals()[f'avd_event_select_{subj_name}'],
                    curr_dct_acts[subj],
                    curr_df_devs
                )
                locals()[f'act_curr_sel_{subj_name}'] = f'avd_graph-event-contingency-{subj}'
                dev_curr_sel = f'avd_graph-event-contingency-{subj}'
                devs_usel_state = False

        #if update_devices:
        if (_is_trigger(ctx, 'devs_graph-bar') or dev_curr_sel == 'devs_graph-bar')\
            and not reset_selection:
            df_dev_curr_sel = _sel_dev_bar(dev_bar_select, df_devs=curr_df_devs)
            dev_curr_sel = 'devs_graph-bar'
            devs_usel_state = False

        if (_is_trigger(ctx, 'devs_graph-fraction') or dev_curr_sel == 'devs_graph-fraction')\
            and not reset_selection:
            df_dev_curr_sel = _sel_dev_fraction(dev_fraction_select, df_devs=curr_df_devs)
            dev_curr_sel = 'devs_graph-fraction'
            devs_usel_state = True

        if (_is_trigger(ctx, 'devs_graph-boxplot') or dev_curr_sel == 'devs_graph-boxplot')\
            and not reset_selection:
            df_dev_curr_sel = _sel_dev_bp_selection(dev_boxplot_select, curr_df_devs)
            dev_curr_sel = 'devs_graph-boxplot'
            devs_usel_state = True

        if (_is_trigger(ctx, 'devs_graph-iei') or dev_curr_sel == 'devs_graph-iei')\
            and not reset_selection:
            df_dev_curr_sel = _sel_dev_iei(dev_iei_select, curr_df_devs,
                                           dev_iei_scale, dev_iei_fig)
            dev_curr_sel = 'devs_graph-iei'
            devs_usel_state = False

        if (_is_trigger(ctx, 'devs_graph-density') or dev_curr_sel == 'devs_graph-density')\
            and not reset_selection:
            df_dev_curr_sel = _sel_dev_density(dev_density_select, curr_df_devs, dev_density_dt)
            dev_curr_sel = 'devs_graph-density'
            devs_usel_state = False



        if reset_selection:
            df_act_curr_sel = None
            df_dev_curr_sel = None
            for subj in dct_acts.subjects():
                locals()[f'act_curr_sel_{subj_name}'] = ''
            dev_curr_sel = ''

        states = (dev_type_trigger == 'state')
        fig_and = activities_and_devices(df_devs1, dct_acts1, st=st, et=et,
                                         states=states, act_order=act_order, dev_order=dev_order,
                                         df_acts_usel=df_act_curr_sel, df_devs_usel=df_dev_curr_sel,
                                         devs_usel_state=devs_usel_state, height=plt_height_ands
                                         )

        # Determine reset selection button values
        if user_sel_activities or user_sel_devices:
            disable_reset = False
        else:
            disable_reset = True

        # Determine if and what updates for activities, device or a~d section are needed
        if data_selection_changed:
            act_update, avd_update, dev_update = 'new_data', 'new_data', 'new_data'
            disable_reset = True
        elif reset_selection:
            act_update, dev_update = "reset_sel", "reset_sel"
            avd_update = dash.no_update
        else:
            if origin_act_bp_selection:
                act_update = 'reset_sel_trans'
            elif origin_act_trans_selection:
                act_update = 'reset_sel_bp'
            elif origin_and_event_selection:
                act_update = 'reset_sel'
            else:
                act_update = dash.no_update

            if origin_dev_density_selection:
                dev_update = 'dens_clicked'
            elif origin_dev_iei_selection:
                dev_update = 'iei_selected'
            elif origin_dev_bp_selection:
                dev_update = 'bp_selected'
            else:
                dev_update = dash.no_update

            # TODO, refactor avd is never updates since it is not necessary
            avd_update = dash.no_update
        try:
            act_curr_sel_store = df_act_curr_sel.to_json(date_unit="ns")
        except:
            act_curr_sel_store = ''
        try:
            dev_curr_sel_store = df_dev_curr_sel.to_json(date_unit="ns"),
        except:
            dev_curr_sel_store = ''

        tmp = []
        for subj_name in dct_acts.subjects():
            tmp += [
                act_update, # TODO update both individually
                locals()[f'act_curr_sel_{subj_name}'],
                locals()[f'act_curr_sel_store_{subj_name}'],
                avd_update
            ]

        ret_arguments = (
               fig_and, act_order, dev_order, disable_reset, 
               dev_update, dev_curr_sel, dev_curr_sel_store,
               *tmp
        )
        return ret_arguments


    @app.callback(
       output=[
            Output('devs_graph-iei', 'figure'),
            Output('devs_graph-density', 'figure'),
            Output('devs_graph-bar', 'figure'),
            Output('devs_graph-fraction', 'figure'),
            Output('devs_graph-boxplot', 'figure'),
            Output('devs_graph-cc', 'figure'),
            Output('devs_order', 'data'),
        ],
        inputs=[
            Input('dev-trigger', 'children'),
            Input('range-slider', 'value'),
            Input('select-devices', 'value'),
            Input('tabs', 'active_tab'),
            Input('devs_bar-scale', 'value'),
            Input('devs_bar-order', 'value'),

            Input('devs_iei-scale', 'value'),
            Input('devs_iei-per-device', 'on'),
            Input('devs_bp-scale', 'value'),
            Input('devs_bp-binary-state', 'value'),
            Input('devs_dens-slider', 'value'),
            Input('devs_dens-scale', 'value'),

            Input('devs_cc-sel-fix', 'value'),
            Input('devs_cc-sel-to', 'value'),
            Input('devs_cc-lag-slider', 'value'),
            Input('devs_cc-binsize-slider', 'value'),
        ],
        state=[
            State('devs_graph-iei', 'figure'),
            State('devs_graph-density', 'figure'),
            State('devs_graph-bar', 'figure'),
            State('devs_graph-fraction', 'figure'),
            State('devs_graph-boxplot', 'figure'),
            State('devs_graph-cc', 'figure'),
            State('devs_order', 'data'),
        ]
    )
    def update_dev_tab(dev_trigger, rng, sel_devices,
                       active_tab, bar_scale, bar_order, iei_scale, iei_per_dev,
                       bp_scale, bp_binary_state, dens_slider, dens_scale, cc_fix, cc_to, cc_slider_lag, cc_slider_bin,
                       fig_iei, fig_dens, fig_left, fig_frac, fig_bp, fig_cc, dev_order, 
                       ):
        ctx = dash.callback_context
        if dash_get_trigger_value(ctx) is None or active_tab != 'tab-devs':
            raise PreventUpdate

        try:
            dev_order = json.loads(dev_order)
        except TypeError:
            dev_order = None

        # Filter selected timeframe, activities and devices
        st = num_to_timestamp(rng[0], start_time=start_time, end_time=end_time)
        et = num_to_timestamp(rng[1], start_time=start_time, end_time=end_time)
        curr_df_devs = select_timespan(df_devs=df_devs, start_time=st, end_time=et,
                                       clip_activities=True)
        curr_df_devs = curr_df_devs[curr_df_devs[DEVICE].isin(sel_devices)]
        from pyadlml.dataset.plot.plotly.devices import bar_count
        from pyadlml.dataset.plot.plotly.devices import fraction as dev_fraction

        is_trigger_bar_drop = _is_trigger(ctx, 'devs_bar-drop')
        is_trigger_bar_scale = _is_trigger(ctx, 'devs_bar-scale')
        is_trigger_bp_scale = _is_trigger(ctx, 'devs_bp-scale')
        is_trigger_bp_state = _is_trigger(ctx, 'devs_bp-binary-state')
        is_trigger_iei_per_device = _is_trigger(ctx, 'devs_iei-per-device')
        is_trigger_iei_scale = _is_trigger(ctx, 'devs_iei-scale')
        is_trigger_dens_scale = _is_trigger(ctx, 'devs_dens-scale')
        is_trigger_dens_slider = _is_trigger(ctx, 'devs_dens-slider')
        is_trigger_sel_dev = _is_trigger(ctx, 'select-devices')
        is_trigger_dev_update = _is_trigger(ctx, 'dev-trigger') or is_trigger_sel_dev

        is_trigger_cc_fix = _is_trigger(ctx, 'devs_cc-sel-fix')
        is_trigger_cc_to = _is_trigger(ctx, 'devs_cc-sel-to')
        is_trigger_cc_slider_lag = _is_trigger(ctx, 'devs_cc-lag-slider')
        is_trigger_cc_slider_bin = _is_trigger(ctx, 'devs_cc-binsize-slider')

        order_update = _is_trigger(ctx, 'devs_bar-order')
        new_data_signal = (is_trigger_dev_update and dev_trigger == 'new_data')

        update_bar = is_trigger_bar_drop or order_update or is_trigger_bar_scale \
                    or new_data_signal
        update_bp = is_trigger_bp_scale or order_update or is_trigger_bp_state \
                    or (is_trigger_dev_update and dev_trigger == 'dens_clicked') \
                    or (is_trigger_dev_update and dev_trigger == 'iei_selected') \
                    or (is_trigger_dev_update and dev_trigger == 'reset_sel') \
                    or new_data_signal
        update_iei = is_trigger_iei_scale or is_trigger_iei_per_device \
                    or new_data_signal
        update_frac = order_update or new_data_signal
        update_density = is_trigger_dens_slider or order_update \
                    or new_data_signal or is_trigger_dens_scale
        update_cc = is_trigger_cc_fix or is_trigger_cc_slider_lag or is_trigger_cc_to \
                    or is_trigger_cc_slider_bin or new_data_signal

        if order_update or dev_order is None or is_trigger_dev_update or is_trigger_sel_dev:
            dev_order = device_order_by(curr_df_devs, rule=bar_order)



        def f_update_frac(fig_frac, cd_dev, order, uf):
            return dev_fraction(cd_dev, order=order, height=plt_height_devs) if uf else fig_frac
        def f_update_bar(ub, fig_left, cdev, scale, order):
            return bar_count(cdev, scale=scale, order=order, height=plt_height_devs) if ub else fig_left
        def f_update_bp(u_bp, fig_bp, cd_dev, order, scale, binary_state):
            if u_bp:
                return boxplot_state(cd_dev, scale=scale, order=order,
                                     binary_state=binary_state, height=plt_height_devs)
            else:
                return fig_bp
        def f_update_density(ud, fig_dens, cd_dev, dens_slider, scale, order):
            if ud:
                return event_density(cd_dev, dt=DEV_DENS_SLIDER[dens_slider], show_colorbar=False,
                                     scale=scale, order=order, height=plt_height_devs)
            else:
                return fig_dens
        def f_update_iei(u_iei, fig_iei, cd_dev, scale, per_dev):
            return dev_iei(cd_dev, scale=scale, per_device=per_dev, height=plt_height_devs) if u_iei else fig_iei

        def f_update_cc(u_cc, fig_cc, fix, to, scale, binsize):
            if u_cc:
                return dev_cc(df_devs, height=plt_height_devs, fix=fix, to=to, max_lag=LAG_SLIDER[scale], binsize=BIN_SIZE_SLIDER[binsize])
            else:
                return fig_cc


        import dask
        figs = [
            dask.delayed(f_update_iei)(update_iei, fig_iei, curr_df_devs, iei_scale, iei_per_dev),
            dask.delayed(f_update_density)(update_density, fig_dens, curr_df_devs,
                                           dens_slider, dens_scale, dev_order),
            dask.delayed(f_update_bar)(update_bar, fig_left, curr_df_devs, bar_scale, dev_order),
            dask.delayed(f_update_frac)(fig_frac, curr_df_devs, dev_order, update_frac),
            dask.delayed(f_update_bp)(update_bp, fig_bp, curr_df_devs, dev_order, bp_scale,
                                      bp_binary_state),
            dask.delayed(f_update_cc)(update_cc, fig_cc, cc_fix, cc_to, cc_slider_lag, cc_slider_bin),
        ]
        figs = dask.compute(figs)[0]

        return figs[0], figs[1], figs[2], figs[3], figs[4], figs[5], json.dumps(list(dev_order))


    for subj_name in dct_acts.keys():
        _create_activity_tab_callback(app, subj_name, dct_acts[subj_name], start_time, end_time, plt_height_acts)
        _create_acts_vs_devs_tab_callback(app, subj_name, dct_acts[subj_name], df_devs, start_time, end_time, dev2area)


def _is_trigger(ctx, val):
    return ctx.triggered[0]['prop_id'].split('.')[0] == val





import pandas as pd
from pathlib import Path
from pyadlml.dataset.io import MegaDownloader
from pyadlml.constants import ACTIVITY, VALUE, \
    START_TIME, END_TIME, TIME, NAME, DEVICE
from pyadlml.dataset._core.devices import device_boolean_on_states_to_events
from pyadlml.dataset.io import DataFetcher

AMSTERDAM_URL = 'https://mega.nz/file/AYhzDLaS#n-CMzBO_raNAgn2Ep1GNgbhah0bHQzuA48PqO_ODEAg'
AMSTERDAM_CLEANED_URL = 'https://mega.nz/file/9R5zxKiT#Ko5NLmAWofmOTGJr0dvjHzjkX8xvKGtZw_Dy5AQ1ogY'
AMSTERDAM_FILENAME = 'amsterdam.zip'
AMSTERDAM_CLEANED_FILENAME = 'amsterdam_cleaned.joblib'
DEVICES_FN = "kasterenSenseData.txt"
ACTIVITIES_FN = "kasterenActData.txt"

__all__ = ['fetch_amsterdam']


class AmsterdamFetcher(DataFetcher):
    def __init__(self):
        downloader = MegaDownloader(
            url=AMSTERDAM_URL,
            fn=AMSTERDAM_FILENAME,
            url_cleaned=AMSTERDAM_CLEANED_URL,
            fn_cleaned=AMSTERDAM_CLEANED_FILENAME,
        )

        super().__init__(
            dataset_name='amsterdam',
            downloader=downloader,
            correct_activities=True,
            correct_devices=True,
        )

    def load_data(self, folder_path):
        device_fp = Path(folder_path).joinpath(DEVICES_FN)
        activity_fp = Path(folder_path).joinpath(ACTIVITIES_FN)

        df_act = _load_activities(activity_fp)
        df_dev = _load_devices(device_fp)
        df_dev = device_boolean_on_states_to_events(df_dev)

        lst_act = df_act[ACTIVITY].unique()
        lst_dev = df_dev[DEVICE].unique()

        return dict(
            activities=df_act,
            devices=df_dev,
            activity_list=lst_act,
            device_list=lst_dev
        )


def fetch_amsterdam(keep_original=False, cache=True, load_cleaned=False,
                    retain_corrections=False, folder_path=None):
    """
    Fetches the amsterdam dataset from the internet. The original dataset or its cached version
    is stored in the data_home folder.

    Parameters
    ----------
    keep_original : bool, default=False
        Determines whether the original dataset is deleted after downloading
        or kept on the hard drive.
    cache : bool, default=True
        Determines whether the data object should be stored as a binary file for quicker access.
        For more information how caching is used refer to the :ref:`user guide <storage>`.
    retain_corrections : bool, default=False
        When set to *true*, data points that change or drop during preprocessing
        are listed in respective attributes of the data object. Fore more information
        about error correction refer to the :ref:`user guide <error_correction>`.
    folder_path : str, default=""
        If set the dataset is loaded from the specified folder.

    Examples
    --------
    >>> from pyadlml.dataset import fetch_amsterdam
    >>> data = fetch_amsterdam()
    >>> data.keys()
    >>> [df_activities, df_devices, activity_list, ... ]

    Returns
    -------
    data : object
    """
    return AmsterdamFetcher()(keep_original=keep_original, cache=cache, load_cleaned=load_cleaned,
                              retain_corrections=retain_corrections, folder_path=folder_path)


def _load_activities(activity_fp):
    act_map = {
        1: 'leave house',
        4: 'use toilet',
        5: 'take shower',
        10: 'go to bed',
        13: 'prepare Breakfast',
        15: 'prepare Dinner',
        17: 'get drink'
    }
    ide = 'id'
    df = pd.read_csv(activity_fp,
                     sep="\t",
                     skiprows=23,
                     skipfooter=1,
                     parse_dates=True,
                     names=[START_TIME, END_TIME, ide],
                     engine='python')
    df[START_TIME] = pd.to_datetime(df[START_TIME])
    df[END_TIME] = pd.to_datetime(df[END_TIME])
    df[ACTIVITY] = df[ide].map(act_map)
    df = df.drop(ide, axis=1)
    return df


def _load_devices(device_fp):
    sens_labels = {
        1: 'Microwave',
        5: 'Hall-Toilet door',
        6: 'Hall-Bathroom door',
        7: 'Cups cupboard',
        8: 'Fridge',
        9: 'Plates cupboard',
        12: 'Frontdoor',
        13: 'Dishwasher',
        14: 'ToiletFlush',
        17: 'Freezer',
        18: 'Pans Cupboard',
        20: 'Washingmachine',
        23: 'Groceries Cupboard',
        24: 'Hall-Bedroom door'
    }
    ide = 'id'
    sens_data = pd.read_csv(device_fp,
                            sep="\t",
                            skiprows=23,
                            skipfooter=1,
                            parse_dates=True,
                            names=[START_TIME, END_TIME, ide, VALUE],
                            engine='python'  # to ignore warning for fallback to python engine because skipfooter
                            # dtype=[]
                            )

    # k todo declare at initialization of dataframe
    sens_data[START_TIME] = pd.to_datetime(sens_data[START_TIME])
    sens_data[END_TIME] = pd.to_datetime(sens_data[END_TIME])
    sens_data[VALUE] = sens_data[VALUE].astype('bool')
    sens_data = sens_data.drop(VALUE, axis=1)

    # replace numbers with the labels
    sens_data[DEVICE] = sens_data[ide].map(sens_labels)
    sens_data = sens_data.drop(ide, axis=1)
    sens_data = sens_data.sort_values(START_TIME)
    return sens_data



from pathlib import Path
from dash import Dash, dash_table, dcc, html
from copy import copy
from dash.dependencies import Input, Output, State
import pandas as pd
import argparse
import dash
from dash.exceptions import PreventUpdate
from pyadlml.constants import ACTIVITY, END_TIME, START_TIME, STRFTIME_PRECISE
from pyadlml.dataset.plot.plotly.util import dash_get_trigger_element, dash_get_trigger_value

from pyadlml.dataset.util import df_difference, fetch_by_name
from pyadlml.dataset.plot.plotly.activities import correction
import dash_bootstrap_components as dbc

"""
for dataset options look under pyadlml.constants:
    casas_aruba
    amsterdam
    mitlab_1
    mitlab_2
    aras
    kasteren_A
    kasteren_B
    kasteren_C
    tuebingen_2019
    uci_ordonezA
    uci_ordonezB
    act_assist
"""

SEP_STR = '#' + '-'*100
ID_STR = '# id='
CASE_STR = 'CaseNr.: %s'
def _df_to_rows(df):
    from pyadlml.constants import STRFTIME_PRECISE
    df = df.copy()
    df[START_TIME] = df[START_TIME].dt.strftime(STRFTIME_PRECISE)
    df[END_TIME] = df[END_TIME].dt.strftime(STRFTIME_PRECISE)
    return df.to_dict('records')
    

def _rows_to_df(rows):
    df = pd.DataFrame(rows, columns=COLS)
    df[START_TIME] = pd.to_datetime(df[START_TIME], errors='coerce', dayfirst=True)
    df[END_TIME] = pd.to_datetime(df[END_TIME], errors='coerce', dayfirst=True)

    # Drop rows with at least one empty cells
    df = df.dropna()
    return df


def delete_id_content_from_code(code, id):
    new_text = []
    text_id = None

    deleting = False
    searching = True
    for line in code.split('\n'):
        if searching:
            if ID_STR in line:
                text_id = int(line[5:])
            if text_id == id:
                searching, deleting = False, True
                continue
            else:
                new_text.append(line)
        elif deleting: 
            if not ID_STR in line and line != SEP_STR:
                continue
            else:
                new_text.append(line)
                deleting = False
        else:
            new_text.append(line)

    code = '\n'.join(new_text)
    return code

def get_id_content_from_code(code, id):
    new_text = []
    text_id = None

    retreiving = False
    searching = True
    for line in code.split('\n'):
        if searching:
            if ID_STR in line:
                text_id = int(line[5:])
            if text_id == id:
                searching, retreiving = False, True
                new_text.append(line)

        elif retreiving: 
            if not ID_STR in line and line != SEP_STR:
                new_text.append(line)
            else:
                break
    code = '\n'.join(new_text)
    return code

def insert_id_content_into_code(code, id, content):
    new_text = code.split('\n')
    text_id = -1
    for idx, line in enumerate(new_text):
        if ID_STR in line:
            text_id = int(line[5:])
        if text_id > id:
            idx += -1
            break
    # Position to insert text
    new_code = '\n'.join(new_text[:idx+1])\
             + content + '\n'.join(new_text[idx+1:]) 
    return new_code

def is_id_in_code(code, id):
    for line in code.split('\n'):
        if ID_STR in line and int(line[5:]) == id:
            return True
    return False

def _update_add_lists_in_footer(content, footer):
    add_lists_names = []
    for line in content.split('\n'):
        if 'lst_to_add_' in line:
            add_lists_names.append(line.split(' ')[0])
    
    new_footer = [] 
    for line in footer.split('\n'):
        if 'new_rows_to_add =' in line:
            line = 'new_rows_to_add = ['
            for lst in add_lists_names:
                line += f'*{lst}, '
            line += ']'
        new_footer.append(line)

    return '\n'.join(new_footer) 

def _update_code(curr_id, text, df_prae, df_post, comment):

    df_diff = df_prae.merge(df_post, indicator=True, how='outer')

    # Rows that are new after correction should be added to the original dataframe
    # and rows that are not present have to be deleted. '_merge'='both' are left alone
    entries_to_add = df_diff[df_diff['_merge'] == 'right_only']
    entries_to_delete = df_diff[df_diff['_merge'] == 'left_only']

    # Delete previous entry if present
    code = delete_id_content_from_code(code=text, id=curr_id)

    # Serialize entries to add 
    content = f'\n{ID_STR}{curr_id}\n'
    content += '\n'.join(['# ' + s for s in comment.split('\n')]) + '\n\n'
    content += f'lst_to_add_{curr_id} = [\n'
    for idx, entry in entries_to_add.iterrows():
        st = entry[START_TIME].strftime(STRFTIME_PRECISE)
        et = entry[END_TIME].strftime(STRFTIME_PRECISE)
        content += f"\t['{st}','{et}','{entry[ACTIVITY]}'],\n"
    content += ']\n'

    # Serialize entries to delete
    content += '\nlst_to_del = [\n'
    for idx, entry in entries_to_delete.iterrows():
        st = entry[START_TIME].strftime(STRFTIME_PRECISE)
        et = entry[END_TIME].strftime(STRFTIME_PRECISE)
        content += f"\t['{st}','{et}','{entry[ACTIVITY]}'],\n"
    content += ']\n'
    content += 'idxs_to_delete += get_index_matching_rows(df_acts, lst_to_del)\n\n'

    # Place content block at appropriate position
    text_head, text_content, text_footer = code.split(SEP_STR)
    text_content = insert_id_content_into_code(text_content, curr_id, content)
    text_footer = _update_add_lists_in_footer(text_content, text_footer)
    text = SEP_STR.join([text_head, text_content, text_footer])

    return text

def df_from_code(code, id, df_acts):
    content_only_from_id = get_id_content_from_code(code, id)
    text_head, _, text_footer = code.split(SEP_STR)
    new_footer = [] 
    # TODO refactor
    for line in text_footer.split('\n'):
        if 'new_rows_to_add =' in line:
            line = f'new_rows_to_add = lst_to_add_{id}'
        new_footer.append(line)

    new_head = []
    for l in text_head.split('\n'):
        if 'df_acts' in l:
            continue
        new_head.append(l)

    code = '\n'.join([*new_head, *content_only_from_id.split('\n'), *new_footer])

    lcls = {'df_acts': df_acts} 
    exec(code, {}, lcls)
    df_acts = lcls['df_acts']
    return df_acts

def comment_from_code(code, id):
    content_only_from_id = get_id_content_from_code(code, id)
    comment = []
    for l in content_only_from_id.split('\n')[1:]:
        if l == '':
            break
        comment.append(l[2:])
    comment = '\n'.join(comment) 
    return comment


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--dataset', type=str, default='amsterdam')
    parser.add_argument('-i', '--identifier', type=str)
    parser.add_argument('-p', '--port', type=str, default='8050')
    parser.add_argument('-dh', '--data-home', type=str, default='/tmp/pyadlml/')
    parser.add_argument('-o', '--out-folder', type=str, default='/tmp/pyadlml/correct_activities')
    args = parser.parse_args()


    out_folder = Path(args.out_folder)
    out_folder.mkdir(parents=True, exist_ok=True)
    fp_code = out_folder.joinpath('corrected_activities.py')

    COLS = [START_TIME, END_TIME, ACTIVITY]
    code_str = '# Awesome script \n'\
                 + 'import pandas as pd\n'\
                 + 'from pyadlml.constants import START_TIME, END_TIME, ACTIVITY\n'\
                 + 'from pyadlml.dataset._core.activities import get_index_matching_rows\n\n'\
                 + '# TODO add original activity dataframe in line below\n'\
                 + 'df_acts = ... \n'\
                 + 'idxs_to_delete = []\n\n'\
                 + '# Content\n' \
                 + SEP_STR +'\n'\
                 + SEP_STR + '\n'\
                 + 'new_rows_to_add = []\n' \
                 + 'df_acts = df_acts.drop(idxs_to_delete)\n'\
                 + "df_news = pd.DataFrame(data=new_rows_to_add, columns=[START_TIME, END_TIME, ACTIVITY])\n"\
                 + "df_news[START_TIME] = pd.to_datetime(df_news[START_TIME], dayfirst=True)\n"\
                 + "df_news[END_TIME] = pd.to_datetime(df_news[END_TIME], dayfirst=True)\n"\
                 + 'df_acts = pd.concat([df_acts, df_news], axis=0)\n'

    with open(fp_code, mode='w') as f:
        f.write(code_str)

    # Setup data
    data = fetch_by_name(args.dataset, identifier=args.identifier, \
                         cache=False, retain_corrections=True, keep_original=True)

    # Create global variables
    corr = data['correction_activities']
    if args.identifier:
        corr = corr[args.identifier]

    app = Dash(__name__, external_stylesheets=[dbc.themes.BOOTSTRAP])

    curr_id = 0
    current_case = corr[curr_id]

    app.layout = dbc.Container([
        dbc.Row([
            dcc.Store(id="curr-id"),
            dcc.Graph(
                id='correction',
                figure=correction(current_case[0], current_case[1])
            )
        ]),
        dbc.Row(dash.html.Header(id='header', children=CASE_STR%(0))),
        dbc.Row([
            dbc.Col([
                dash_table.DataTable(
                    id='pre-data-table',
                    columns=(
                        [{'id':c, 'name' :c } for c in current_case[0].columns]
                    ),
                    data=_df_to_rows(current_case[0]),
                    editable=False,
                    style_table={'height': '200px', 'overflowY': 'auto'},
                    style_cell = {'font-size': '12px'},
                ),
            ], md=6),
            dbc.Col([
                dbc.Row(
                    dash_table.DataTable(
                        id='post-data-table',
                        columns=(
                            [{'id':c, 'name' :c } for c in current_case[0].columns]
                        ),
                        data=_df_to_rows(current_case[1]),
                        editable=True,
                        row_deletable=True,
                        style_table={'height': '200px', 'overflowY': 'auto'},
                        style_cell = {'font-size': '12px'},
                    ),
                ),
                dbc.Row([
                    dbc.ButtonGroup([
                        dbc.Button(id="reset", children="reset", color="secondary"), 
                        dbc.Button(id="reset_heuristic", children="reset to heuristic", color="secondary"), 
                        dbc.Button(id="add", children="add empty row", color="secondary"), 
                    ])
                ])
            ], md=6)
        ], className='mb-1'),
        dbc.Row(
            dbc.Textarea(id='code', value='', readOnly=False, 
                        style={'height': '100px'}
                        #style={'resize': 'none', 'overflow':'hidden'}#, 'height': '100px'}
            ),
            style={'height':'100%'},
            className='mb-1',
        ),
        dbc.Row([
            dbc.ButtonGroup([
                dbc.Button(id="prev", children="Previous", color="secondary"), 
                dbc.Button(id="submit", children="Submit", color="primary"),            
                dbc.Button(id="next", children="Next", color="secondary"), 
            ])
        ], className='mb-5'),

    ], style={'width': 1500, 'margin': 'auto'})

    @app.callback(
        Output('correction', 'figure'),
        Output('post-data-table', 'data'),
        Output('pre-data-table', 'data'),
        Output('code', 'value'),
        Output('curr-id', 'data'),
        Output('header', 'children'),
        Output('submit', 'children'),
        Input('post-data-table', 'data'),
        Input('add', 'n_clicks'),
        Input('reset', 'n_clicks'),
        Input('reset_heuristic', 'n_clicks'),
        Input('prev', 'n_clicks'),
        Input('submit', 'n_clicks'),
        Input('next', 'n_clicks'),
        State('pre-data-table', 'data'),
        State('correction', 'figure'),
        State('code', 'value'),
        State('curr-id', 'data'),
        State('submit', 'children'),
    )
    def display_output(post_rows, add_click, reset_click, reset_heuristic_click, prev_click,
                      sub_click, next_click, prae_rows, fig, comment, curr_id, button_text
    ):
        curr_id = 0 if curr_id is None else curr_id

        current_case = corr[curr_id]

        trigger = dash_get_trigger_element()
        if trigger is None or trigger == '':
            raise PreventUpdate


        # Open code
        f = open(fp_code, mode='r')
        code_str = ''.join(f.readlines())
        f.close()


        if trigger == 'add' and add_click > 0: 
            post_rows.append({c:'' for c in COLS})

        elif trigger == 'reset':
            fig = correction(current_case[0], current_case[0])
            post_rows = _df_to_rows(current_case[0])

        elif trigger == 'reset_heuristic':
            fig = correction(current_case[0], current_case[1])
            post_rows = _df_to_rows(current_case[1])

        elif trigger == 'post-data-table':
            df_post = _rows_to_df(post_rows)
            fig = correction(current_case[0], df_post)

        elif trigger == 'submit':
            df_post = _rows_to_df(post_rows)
            code_str = _update_code(curr_id, code_str, current_case[0], df_post, comment)

        elif trigger == 'prev' or trigger in 'next':
            # Assign new case
            curr_id = max(curr_id-1,0) if trigger == 'prev' else min(curr_id+1, len(corr)-1)
            current_case = corr[curr_id]

            if is_id_in_code(code_str, curr_id): 
                comment = comment_from_code(code_str, curr_id)
                df = df_from_code(code_str, curr_id, copy(current_case[0]))
                post_rows = _df_to_rows(df)
                df_post = df
            else:
                comment = ''
                post_rows = _df_to_rows(current_case[1])
                df_post = current_case[1]

            # Update stuff
            fig = correction(current_case[0], df_post)
            prae_rows = _df_to_rows(current_case[0])

        else:
            raise ValueError


        button_text = 're-submit' if is_id_in_code(code_str, curr_id) else 'submit'  

        with open(fp_code, mode='w') as f:
            f.write(code_str)

        return fig, post_rows, prae_rows, comment, curr_id, CASE_STR%(curr_id), button_text

    app.run_server(debug=False)



import sys
import pathlib
working_directory = pathlib.Path().absolute()
script_directory = pathlib.Path(__file__).parent.absolute()
sys.path.append(str(working_directory))
import unittest


class TestImport(unittest.TestCase):
    def setUp(self):
        import pyadlml
        
    def test_import_datasets(self):
        from pyadlml.dataset import load_act_assist
        from pyadlml.dataset import fetch_amsterdam
        from pyadlml.dataset import fetch_aras
        from pyadlml.dataset import fetch_casas_aruba
        from pyadlml.dataset import fetch_mitlab
        from pyadlml.dataset import fetch_tuebingen_2019
        from pyadlml.dataset import fetch_uci_adl_binary
    
    def test_import_plots(self):
        # import activity plots
        from pyadlml.plot import plot_activity_bp_duration, plot_activity_bar_duration, \
            plot_activity_bar_count, plot_activity_hm_transition, plot_activity_rl_daily_density

        # import device plots
        from pyadlml.plot import plot_device_on_off, plot_device_bar_count, plot_device_hm_similarity,\
            plot_device_bp_on_duration, plot_device_hm_time_trigger, plot_device_hist_time_diff_trigger

        # import device and activity interaction
        from pyadlml.plot import plot_hm_contingency_trigger, plot_hm_contingency_trigger_01,  \
            plot_hm_contingency_duration

    def test_import_stats(self):
        from pyadlml.stats import activity_dist, activity_count, activity_duration, activity_duration_dist, \
            activity_transition

        from pyadlml.stats import device_duration_corr, device_on_off, device_on_time, device_time_diff, \
            device_trigger_count, device_trigger_one_day, device_trigger_sliding_window

        from pyadlml.stats import contingency_duration, contingency_triggers_01, contingency_triggers

if __name__ == '__main__':
    unittest.main()